//****************************************************
//
//	project	PAFEE
//!	@file:	ArrayT.h
//!	@author: 
//!	@brief:	
//
//****************************************************
//
// P A F E E   - Pragmatic Application Framework for Embedded Environment- 
//
//
// Copyright (c) 2009 KN Information Systems Corporation,
//                    Hirocon Corporation,
//                    Technocrats Corporation
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//

#ifndef _ARRAY_T_H_INCLUDED_
#define _ARRAY_T_H_INCLUDED_

#include "PafeeCommon.h"
#include "PafeeObject.h"


#ifdef PAFEE_UNITTEST
#include <assert.h>
#endif // PAFEE_UNITTEST

namespace Pafee
{
	template<class T>
	class ArrayT : public Object
	{
	public:
		ArrayT(int initialCapacity=0, int extendCapacity=8, Allocator* pAllocator=NULL)
		{
			m_isInitialized = false;
			Initialize(initialCapacity, extendCapacity, pAllocator);
		};

		~ArrayT()
		{
			if(m_pElementArray)
			{
				delete[] m_pElementArray;
			}
		};
	public:
#ifdef PAFEE_STATISTICS_ENABLE
		typedef struct
		{
			int addCount;
			int insertCount;
			int removeCount;
			int currentElementCount;
		} STATISTICS;
#endif // PAFEE_STATISTICS_ENABLE
	public:

		bool Initialize(int initialCapacity, int extendCapacity, Allocator* pAllocator=NULL)
		{
			// reinitialize is ignored
			if(!m_isInitialized)
			{
				m_pElementArray = NULL;

#ifdef PAFEE_STATISTICS_ENABLE
				m_statistics.addCount = 0;
				m_statistics.insertCount = 0;
				m_statistics.removeCount = 0;
#endif // PAFEE_STATISTICS_ENABLE

				m_pAllocator = pAllocator;

				m_arrayCapacity = initialCapacity;

				if(m_arrayCapacity)
				{
					m_pElementArray = new T[m_arrayCapacity];
					if(!m_pElementArray)
					{
						/* @@@throw exception */
						return false;
					}
				}

				m_extendCapacity = extendCapacity;

				m_elementCount = m_arrayCapacity;

				m_isInitialized = true;
			}
#ifdef PAFEE_UNITTEST
			else
			{
				return false;
			}
#endif // PAFEE_UNITTEST
			return true;
		}


		int GetCount(void)
		{
			return m_elementCount;
		}


		bool IsEmpty(void)
		{
			return (m_elementCount)? false: true;
		}


		int GetCapacity(void)
		{
			return m_arrayCapacity;
		}


		void SetCapacity(int capacity)
		{
			if(capacity != m_arrayCapacity && capacity > 0)
			{
				// reallocate
				ReallocateElementArray(capacity);

				m_elementCount = m_arrayCapacity;
			}
		}


		T& GetAt(int index)
		{
#ifdef PAFEE_UNITTEST
			assert(index < m_elementCount);
#endif // PAFEE_UNITTEST
			return m_pElementArray[index];
		}


		void SetAt(int index, const T& data)
		{
#ifdef PAFEE_UNITTEST
			assert(index < m_elementCount);
#endif // PAFEE_UNITTEST
			m_pElementArray[index] = data;
		}


		void Add(const T& data)
		{
#ifdef PAFEE_STATISTICS_ENABLE
			m_statistics.addCount++;
#endif // PAFEE_STATISTICS_ENABLE
			if(m_elementCount == m_arrayCapacity)
			{
				ReallocateElementArray(m_arrayCapacity + m_extendCapacity);
			}

			// add element
			m_elementCount++;
			SetAt(m_elementCount-1, data);
		}


		void InsertAt(int index, const T& data)
		{
			if(index <= m_elementCount)
			{
#ifdef PAFEE_STATISTICS_ENABLE
				m_statistics.insertCount++;
#endif // PAFEE_STATISTICS_ENABLE

				// check if arrayCapacity is shortage
				if(m_elementCount == m_arrayCapacity)
				{
					// reallocate array
					ReallocateElementArray(m_arrayCapacity + m_extendCapacity);
				}

				// shift old elements
				MoveElement(&m_pElementArray[index+1], &m_pElementArray[index], (m_elementCount - index));

				// update element count
				m_elementCount++;

				SetAt(index, data);
			}
#ifdef PAFEE_UNITTEST
			else
			{
				// @@@output log
				assert(0);
			}
#endif // PAFEE_UNITTEST
		}


		void RemoveAt(int index)
		{
			if(index < m_elementCount)
			{
				// shift old elements
				MoveElement(&m_pElementArray[index], &m_pElementArray[index+1], (m_elementCount - (index+1)));

				// update element count
				m_elementCount--;
			}
#ifdef PAFEE_UNITTEST
			else
			{
				// @@@output log
				assert(0);
			}
#endif // PAFEE_UNITTEST
		}


		void RemoveAll(void)
		{
#ifdef PAFEE_STATISTICS_ENABLE
			m_statistics.removeCount++;
#endif // PAFEE_STATISTICS_ENABLE
			m_elementCount = 0;
		}


		T& operator[] (int index) { return m_pElementArray[index]; }

#ifdef PAFEE_STATISTICS_ENABLE
		void GetStatistics(ArrayT::STATISTICS* pStatistics)
		{
			pStatistics->addCount = m_statistics.addCount;
			pStatistics->insertCount = m_statistics.insertCount;
			pStatistics->removeCount = m_statistics.removeCount;
			pStatistics->currentElementCount = GetCount();
		}
#endif // PAFEE_STATISTICS_ENABLE


	private:

		bool ReallocateElementArray(int newCapacity)
		{
			if(m_arrayCapacity != newCapacity)
			{
				int i;
				int capacity;
				T* pNewElementArray = NULL;
				T* pNewElement;
				T* pOldElement;

				pNewElementArray = new T[newCapacity];

				if(!pNewElementArray)
				{
					return false;
				}

				// copy element array
				pNewElement = pNewElementArray;
				pOldElement = m_pElementArray;

				if(m_arrayCapacity < newCapacity)
				{
					capacity = m_arrayCapacity;
				}
				else // (oldCapacity > newCapacity)
				{
					capacity = newCapacity;
				}

				for(i=0; i<capacity; i++)
				{
					*pNewElement++ = *pOldElement++;
				}

				// free old element array
				if(m_pElementArray)
				{
					delete[] m_pElementArray;
				}

				// set new element array
				m_pElementArray = pNewElementArray;

				// update element array capacity
				m_arrayCapacity = newCapacity;
			}
			return true;
		}


		void MoveElement(T* destination, T* source, int numberOfObjects)
		{
			int i;

			if(destination > source)
			{
				T* newData = destination + numberOfObjects - 1;
				T* oldData = source + numberOfObjects - 1;
				// @@@should use optimized copy function
				for(i=0; i<numberOfObjects; i++)
				{
					*newData-- = *oldData--;
				}
			}
			else if(destination < source)
			{
				T* newData = destination;
				T* oldData = source;
				// @@@should use optimized copy function
				for(i=0; i<numberOfObjects; i++)
				{
					*newData++ = *oldData++;
				}
			}
			/* else if(destination == source) -> ignored */
		}

#ifdef PAFEE_UNITTEST
public:
		static void UnitTest(void)
		{
#define TEST_ARRAY_T_CAPACITY	10
#define TEST_ARRAY_T_EXTEND_CAPACITY	8
			ArrayT<char>	array(TEST_ARRAY_T_CAPACITY, TEST_ARRAY_T_EXTEND_CAPACITY, NULL);
			char			pData[TEST_ARRAY_T_CAPACITY*10];
			int			i,j,t;
			int			addCount = 0;
			int			insertCount = 0;
			int			removeCount = 0;
#ifdef PAFEE_STATISTICS_ENABLE
			STATISTICS		statistics;
#endif // PAFEE_STATISTICS_ENABLE

			// Initialize test data
			{
				for(i=0; i<TEST_ARRAY_T_CAPACITY*10; i++)
				{
					pData[i] = i;
				}
			}

			//
			// test Initialize/GetCount/GetCapacity/IsEmpty
			//
			{
				assert(!array.Initialize(TEST_ARRAY_T_CAPACITY, TEST_ARRAY_T_EXTEND_CAPACITY, NULL));
				assert(array.GetCount() == TEST_ARRAY_T_CAPACITY);
				assert(array.GetCapacity() == TEST_ARRAY_T_CAPACITY);
			}

			//
			// test GetAt/SetAt/GetCount/operator[]
			//
			{
				// SetAt/GetAt
				for(i=0; i<TEST_ARRAY_T_CAPACITY; i++)
				{
					array.SetAt(i, 0);
				}
				for(i=0; i<TEST_ARRAY_T_CAPACITY; i++)
				{
					assert(array.GetAt(i) == 0);
				}
				for(i=0; i<TEST_ARRAY_T_CAPACITY; i++)
				{
					array.SetAt(i, i);
				}
				for(i=0; i<TEST_ARRAY_T_CAPACITY; i++)
				{
					assert(array.GetAt(i) == i);
				}

				// operator[]
				for(i=0; i<TEST_ARRAY_T_CAPACITY; i++)
				{
					array[i] = 0;
				}
				for(i=0; i<TEST_ARRAY_T_CAPACITY; i++)
				{
					assert(array[i] == 0);
				}
				for(i=0; i<TEST_ARRAY_T_CAPACITY; i++)
				{
					array[i] = i;
				}
				for(i=0; i<TEST_ARRAY_T_CAPACITY; i++)
				{
					assert(array[i] == i);
				}
			}

			//
			// test SetCapacity
			//
			{
				char* pOldPointer = &array[0];
				array.SetCapacity(TEST_ARRAY_T_CAPACITY);
				assert(&array[0] == pOldPointer);
				for(i=0; i<TEST_ARRAY_T_CAPACITY; i++)
				{
					array[i] = i;
				}

				for(i=0; i<TEST_ARRAY_T_CAPACITY-1; i++)
				{
					pOldPointer = &array[0];
					array.SetCapacity(TEST_ARRAY_T_CAPACITY - i - 1);
					assert((&array[0] != NULL) && (&array[0] != pOldPointer));
					assert(array.GetCount() == TEST_ARRAY_T_CAPACITY - i - 1);
					assert(array.GetCapacity() == TEST_ARRAY_T_CAPACITY - i - 1);
					for(j=0; j<(TEST_ARRAY_T_CAPACITY - i - 1); j++)
					{
						assert(array[j] == j);
					}
				}

				pOldPointer = &array[0];
				array.SetCapacity(0);
				assert((&array[0] != NULL) && (&array[0] == pOldPointer));
				assert(array.GetCount() == 1);
				assert(array.GetCapacity() == 1);


				for(i=1; i<TEST_ARRAY_T_CAPACITY; i++)
				{
					pOldPointer = &array[0];
					array.SetCapacity(i+1);
					assert((&array[0] != NULL) && (&array[0] != pOldPointer));
					assert(array.GetCount() == i+1);
					assert(array.GetCapacity() == i+1);
					for(j=0; j<i; j++)
					{
						array[j] = j;
					}
					for(j=0; j<i; j++)
					{
						assert(array[j] == j);
					}
				}
			}

			//
			// test Add
			//
			{
				char*	pOldPointer;
				int	oldCapacity;
				int	count = TEST_ARRAY_T_CAPACITY;

				for(i=0; i<TEST_ARRAY_T_CAPACITY; i++)
				{
					array[i] = pData[i];
				}

				for(i=0; i<5; i++)
				{
					for(j=0; j<TEST_ARRAY_T_EXTEND_CAPACITY; j++)
					{
						pOldPointer = &array[0];
						oldCapacity = array.GetCapacity();

						array.Add(pData[count++]);
						addCount++;

						if(j==0)
						{
							assert(&array[0] != pOldPointer);
							assert(array.GetCapacity() == (oldCapacity + TEST_ARRAY_T_EXTEND_CAPACITY));
						}
						else
						{
							assert(&array[0] == pOldPointer);
							assert(array.GetCapacity() == oldCapacity);
						}
						assert(array.GetCount() == count);
						for(t=0; t<count; t++)
						{
							assert(array[t] == pData[t]);
						}
					}
				}
			}

			//
			// test InsertAt/RemoveAt
			//
			{
				char*	pOldPointer;
				int	oldCapacity;
				int	count = TEST_ARRAY_T_CAPACITY + TEST_ARRAY_T_EXTEND_CAPACITY*5;

				// RemoveAt
				for(i=0; i<TEST_ARRAY_T_EXTEND_CAPACITY*5; i++)
				{
					array.RemoveAt(0);
					removeCount++;
					count--;
				}
				assert(array.GetCount() == count);
				for(i=0; i<count; i++)
				{
					assert(array[i] == pData[TEST_ARRAY_T_EXTEND_CAPACITY*5+i]);
				}
				array.SetCapacity(TEST_ARRAY_T_CAPACITY);

				// InsertAt
				for(i=0; i<5; i++)
				{
					for(j=0; j<TEST_ARRAY_T_EXTEND_CAPACITY; j++)
					{
						pOldPointer = &array[0];
						oldCapacity = array.GetCapacity();

						int aaaaa = 5*TEST_ARRAY_T_EXTEND_CAPACITY + TEST_ARRAY_T_CAPACITY - count - 1;

						array.InsertAt(0, pData[5*TEST_ARRAY_T_EXTEND_CAPACITY + TEST_ARRAY_T_CAPACITY - count - 1]);
						insertCount++;
						count++;

						if(j==0)
						{
							assert(&array[0] != pOldPointer);
							assert(array.GetCapacity() == oldCapacity + TEST_ARRAY_T_EXTEND_CAPACITY);
						}
						else
						{
							assert(&array[0] == pOldPointer);
							assert(array.GetCapacity() == oldCapacity);
						}
						assert(array.GetCount() == count);
					}
				}
				for(i=0; i<count; i++)
				{
					assert(array[i] == pData[i]);
				}

				//
				for(i=0; i<count; i++)
				{
					array.RemoveAt(i);
					removeCount++;
					array.InsertAt(i, pData[count - i - 1]);
					insertCount++;
				}
				assert(array.GetCount() == count);
				for(i=0; i<count; i++)
				{
					assert(array[i] == pData[count - i - 1]);
				}
			}

			//
			// test RemoveAll
			//
			{
				char*	pOldPointer = &array[0];
				int	oldCapacity = array.GetCapacity();
				int	count = array.GetCount();

				array.RemoveAll();

				assert(&array[0] == pOldPointer);
				assert(array.GetCapacity() == oldCapacity);
				assert(array.GetCount() == 0);

				for(i=0; i<count; i++)
				{
					array.Add(pData[i]);
					addCount++;
					assert(&array[0] == pOldPointer);
					assert(array.GetCapacity() == oldCapacity);
					assert(array.GetCount() == (i+1));
				}
			}

#ifdef PAFEE_STATISTICS_ENABLE
			//
			// GetStatistics
			//
			{
				array.GetStatistics(&statistics);
				assert(statistics.addCount == addCount);
				assert(statistics.insertCount == insertCount);
				assert(statistics.currentElementCount == array.GetCount());
			}
#endif // PAFEE_STATISTICS_ENABLE
		}
#endif // PAFEE_UNITTEST


	private:

		T* m_pElementArray;

		int m_arrayCapacity;	// number of elements(current memory capacity that is allocated)

		int m_elementCount;		// number of elements(current number of data that is set by user)

		int m_extendCapacity;	// number of elements(memory capacity that is added when current memory capacity is shortage)

		Allocator* m_pAllocator;

		bool m_isInitialized;

#ifdef PAFEE_STATISTICS_ENABLE
		STATISTICS m_statistics;
#endif // PAFEE_STATISTICS_ENABLE

	};

};	// Pafee

#endif // _ARRAY_T_H_INCLUDED_
