//****************************************************
//
//    project   Pafee
//!   @file:    PafeeList.cpp
//!   @author:  M.Tanaka
//!   @brief:   List Class
//
//*****************************************************

//
// P A F E E   - Pragmatic Application Framework for Embedded Environment- 
//
//
// Copyright (c) 2009 KN Information Systems Corporation,
//                    Hirocon Corporation,
//                    Technocrats Corporation
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//

#include "PafeeList.h"

namespace Pafee
{

//----------------------------------------
//! Default Constructor
//----------------------------------------
List::List(
	Allocator* pAllocator ///< [in] Memory Allocator
)
{
	Initialize(pAllocator);
}

//----------------------------------------
//! Copy Constructor
//----------------------------------------
List::List(
	const List& lstc, ///< [in] initial value of List
	Allocator* pAllocator ///< [in] Memory Allocator
)
{
	Initialize(pAllocator);
	List::POSITION pFind = lstc.m_pStart;
	while(pFind != NULL)
	{
		this->AddBack(pFind->m_pData,pFind->m_dateLength);
		pFind = pFind->m_pNext;
	}
	if(lstc.m_pWriteBox != NULL)
	{
		char* p = (char*)this->GetWriteBuffer(lstc.m_pWriteBox->m_bufferSize);
		memcpy(p,lstc.m_pWriteBox->m_pData,lstc.m_pWriteBox->m_bufferSize);
	}
	STATISTICS_SET(m_Statistics.addCount,            lstc.m_Statistics.addCount);
	STATISTICS_SET(m_Statistics.deleteCount,         lstc.m_Statistics.deleteCount);
	STATISTICS_SET(m_Statistics.maxElementCount,     lstc.m_Statistics.maxElementCount);
	STATISTICS_SET(m_Statistics.maxElementSize,      lstc.m_Statistics.maxElementSize);
	STATISTICS_SET(m_Statistics.minElementSize,      lstc.m_Statistics.minElementSize);
	STATISTICS_SET(m_Statistics.presentElementCount, lstc.m_Statistics.presentElementCount);
}

//----------------------------------------
//! Destructor
//----------------------------------------
List::~List()
{
	List::POSITION p;
	List::POSITION pdel;
	//
	if(m_pWriteBox != NULL)
	{
		Deallocate( (char* )m_pWriteBox);
	}
	for(p=m_pEnd;p!=NULL;)
	{
		pdel = p;
		p = p->m_pPrev;
		DestroyBox(pdel);
	}
}

//----------------------------------------
//! Initialize
//! @return: Result(ture)
//----------------------------------------
bool List::Initialize(
	Allocator* pAllocator ///< [in]  Memory Allocator
)
{
	SetAllocator(pAllocator);
	m_pStart = NULL;
	m_pEnd = NULL;
	m_pWriteBox = NULL;
	m_elementNum = 0;
	STATISTICS_SET(m_Statistics.addCount,            0);
	STATISTICS_SET(m_Statistics.deleteCount,         0);
	STATISTICS_SET(m_Statistics.maxElementCount,     0);
	STATISTICS_SET(m_Statistics.maxElementSize,      0);
	STATISTICS_SET(m_Statistics.minElementSize,      0);
	STATISTICS_SET(m_Statistics.presentElementCount, 0);
	return true;
}

//-------------------------------------------------------------------------------------
//! operator =
//! @return Result(the address of List)
//-------------------------------------------------------------------------------------
List& List::operator =(
	const List& lstc ///< [in] Object of copied list
)
{
	this->RemoveAll();
	if(m_pWriteBox != NULL)
	{
		Deallocate( (char* )m_pWriteBox);
		m_pWriteBox = NULL;
	}
	List::POSITION pFind = lstc.m_pStart;
	while(pFind != NULL)
	{
		this->AddBack(pFind->m_pData,pFind->m_dateLength);
		pFind = pFind->m_pNext;
	}
	if(lstc.m_pWriteBox != NULL)
	{
		char* p = (char*)this->GetWriteBuffer(lstc.m_pWriteBox->m_bufferSize);
		memcpy(p,lstc.m_pWriteBox->m_pData,lstc.m_pWriteBox->m_bufferSize);
	}
	STATISTICS_SET(m_Statistics.addCount,            lstc.m_Statistics.addCount);
	STATISTICS_SET(m_Statistics.deleteCount,         lstc.m_Statistics.deleteCount);
	STATISTICS_SET(m_Statistics.maxElementCount,     lstc.m_Statistics.maxElementCount);
	STATISTICS_SET(m_Statistics.maxElementSize,      lstc.m_Statistics.maxElementSize);
	STATISTICS_SET(m_Statistics.minElementSize,      lstc.m_Statistics.minElementSize);
	STATISTICS_SET(m_Statistics.presentElementCount, lstc.m_Statistics.presentElementCount);
	return *this;
}

//----------------------------------------
//! IsEmpty
//! @return: Result(ture:node empty false:node exist)
//----------------------------------------
bool List::IsEmpty()
{
	if(m_elementNum)
	{
		return false;
	}
	return true;
}

//----------------------------------------
//! GetWriteBuffer
//! @return: Result(New Buffer Pointer)
//----------------------------------------
void* List::GetWriteBuffer(
	int length ///< [in] Buffer Size
)
{
	m_pWriteBox = CreateBox(NULL,length);
	if ( m_pWriteBox == NULL )
	{
		return NULL;
	}
	return m_pWriteBox->m_pData;
}

//----------------------------------------
//! AddFront
//! @return: Result(true:Success false:failure)
//----------------------------------------
bool List::AddFront(
	const void* pData, ///< [in] Write Thing 
	int length ///< [in] Write Size
)
{
	List::POSITION pNew;
	//
	pNew = CreateBox(pData,length);
	if(pNew == NULL)
	{
		return false;
	}
	if(m_pStart == NULL)
	{
		m_pStart = m_pEnd = pNew;
	}
	else
	{
		pNew->m_pNext = m_pStart;
		m_pStart->m_pPrev = pNew;
		m_pStart = pNew;
	}
	m_elementNum++;
	//
	STATISTICS_INC(m_Statistics.addCount);
	STATISTICS_INC(m_Statistics.presentElementCount);
	STATISTICS_MAX(m_Statistics.maxElementCount,m_Statistics.presentElementCount);
	//
	return true;
}

//----------------------------------------
//! AddFront
//! @return: Result(true:Success false:failure)
//----------------------------------------
bool List::AddFront(
	int length ///< [in] data size
)
{
	if(m_pWriteBox == NULL)
	{
		return false;
	}
	m_pWriteBox->m_dateLength = length;

	if(m_pStart == NULL)
	{
		m_pStart = m_pEnd = m_pWriteBox;
	}
	else
	{
		m_pWriteBox->m_pNext = m_pStart;
		m_pStart->m_pPrev = m_pWriteBox;
		m_pStart = m_pWriteBox;
	}
	m_elementNum++;
	m_pWriteBox = NULL;
	//
	STATISTICS_INC(m_Statistics.addCount);
	STATISTICS_INC(m_Statistics.presentElementCount);
	STATISTICS_MAX(m_Statistics.maxElementCount,m_Statistics.presentElementCount);
	//
	return true;
}

//----------------------------------------
//! AddBack
//! @return: Result(true:Success false:failure)
//----------------------------------------
bool List::AddBack(
	const void *pData, ///< [in]Write Thing 
	int length       ///< [in]Write Size
)
{
	List::POSITION pNew;
	pNew = CreateBox(pData,length);
	if(pNew == NULL)
	{
		return false;
	}
	if(m_pStart == NULL)
	{
		m_pStart = m_pEnd = pNew;
	}
	else
	{
		pNew->m_pPrev = m_pEnd;
		m_pEnd->m_pNext = pNew;
		m_pEnd = pNew;
	}
	m_elementNum++;
	//
	STATISTICS_INC(m_Statistics.addCount);
	STATISTICS_INC(m_Statistics.presentElementCount);
	STATISTICS_MAX(m_Statistics.maxElementCount,m_Statistics.presentElementCount);
	//
	return true;
}

//----------------------------------------
//! AddBack
//! @return: Result(true:Success false:failure)
//----------------------------------------
bool List::AddBack(
	int length ///< [in] data size
)
{
	if(m_pWriteBox == NULL)
	{
		return false;
	}
	m_pWriteBox->m_dateLength = length;
	if(m_pStart == NULL)
	{
		m_pStart = m_pEnd = m_pWriteBox;
	}
	else
	{
		m_pWriteBox->m_pPrev = m_pEnd;
		m_pEnd->m_pNext = m_pWriteBox;
		m_pEnd = m_pWriteBox;
	}
	m_elementNum++;
	m_pWriteBox = NULL;
	//
	STATISTICS_INC(m_Statistics.addCount);
	STATISTICS_INC(m_Statistics.presentElementCount);
	STATISTICS_MAX(m_Statistics.maxElementCount,m_Statistics.presentElementCount);
	//
	return true;
}

//----------------------------------------
//! GetFrontLength
//! @return: Result(Size of a head element)
//----------------------------------------
int List::GetFrontLength()
{
	if(m_pStart == NULL)
	{
		return 0;
	}
	return m_pStart->m_dateLength;
}

//----------------------------------------
//! GetBackLength
//! @return: Result(Size of a end element)
//----------------------------------------
int List::GetBackLength()
{
	if(m_pEnd == NULL)
	{
		return 0;
	}
	return m_pEnd->m_dateLength;
}

//----------------------------------------
//! GetLengthAt
//! @return: Result(Size of a specification element)
//----------------------------------------
int List::GetLengthAt(
	List::POSITION pos ///< [in] Specified Position
)
{
	if(pos != NULL)
	{
		return pos->m_dateLength;
	}
	return 0;
}

//----------------------------------------
//! GetFront
//! @return: Result(true:Success false:failure)
//----------------------------------------
bool List::GetFront(
	void* pBuffer, ///< [out] Head Element 
	int* pLength, ///< [out] Head Element Size
	int bufSize ///< [in] Prepared buffer sizes
)
{
	if(m_pStart == NULL)
	{
		return false;
	}
	if(bufSize < m_pStart->m_dateLength)
	{
		return false;
	}
	pafee_memcpy(pBuffer,m_pStart->m_pData,m_pStart->m_dateLength);
	*pLength = m_pStart->m_dateLength;
	return true;
}

//----------------------------------------
//! GetFront
//! @return: Result(true:Success false:failure)
//----------------------------------------
void* List::GetFront(
	int* pLength ///< [out] Head Element Size
)
{
	if(m_pStart == NULL)
	{
		return NULL;
	}
	*pLength = m_pStart->m_dateLength;
	return m_pStart->m_pData;
}

//----------------------------------------
//! GetBack
//! @return: Result(true:Success false:failure)
//----------------------------------------
bool List::GetBack(
	void* pBuffer, ///< [out] End Element
	int* pLength, ///< [out] End Element Size
	int bufSize ///< [in] Prepared buffer sizes
)
{
	if(m_pStart == NULL)
	{
		return false;
	}
	if(bufSize < m_pEnd->m_dateLength)
	{
		return false;
	}
	pafee_memcpy(pBuffer,m_pEnd->m_pData,m_pEnd->m_dateLength);
	*pLength = m_pEnd->m_dateLength;
	return true; 
}

//----------------------------------------
//! GetBack
//! @return: Result(true:Success false:failure)
//----------------------------------------
void* List::GetBack(
	int* pLength ///< [out] End Element Size
)
{
	if(m_pStart == NULL)
	{
		return NULL;
	}
	*pLength = m_pEnd->m_dateLength;
	return m_pEnd->m_pData;
}

//----------------------------------------
//! GetNextLength
//! @return: Result(Next Data Length)
//----------------------------------------
int List::GetNextLength(
	List::POSITION& pos ///< [in] Specified Position
)
{
	if(pos != NULL)
	{
		if(pos->m_pNext != NULL)
		{
			return pos->m_pNext->m_dateLength;
		}
	}
	return 0;
}

//----------------------------------------
//! GetPreviousLength
//! @return: Result(Previous Data Length)
//----------------------------------------
int List::GetPreviousLength(
	List::POSITION& pos ///< [in] Specified Position
)
{
	if(pos != NULL)
	{
		if(pos->m_pPrev != NULL)
		{
			return pos->m_pPrev->m_dateLength;
		}
	}
	return 0;
}

//----------------------------------------
//! GetNext
//! @return: Result(true:Success false:failure)
//----------------------------------------
bool List::GetNext(
	List::POSITION &pos, ///< [in/out] Next Node Pointer
	void* pBuffer, ///< [out] Spesified Position Element  
	int* pLength, ///< [out] Spesified Position Element Size
	int bufSize ///< [in] Prepared buffer sizes 
)
{
	List::POSITION pTarget = pos;
	if(pTarget != NULL)
	{
		if(pTarget->m_pData != NULL){
			if(bufSize < pTarget->m_dateLength)
			{
				return false;
			}
			pos = (POSITION) pTarget->m_pNext;
			pafee_memcpy(pBuffer,pTarget->m_pData,pTarget->m_dateLength);
			*pLength = pTarget->m_dateLength;
			return true;
		}
	}
	return false;
}

//----------------------------------------
//! GetNext
//! @return: Result(The element pointer of the specified position)
//----------------------------------------
void* List::GetNext(
	List::POSITION &pos, ///< [in/out] Next Node Pointer
	int* pLength ///< [out] Spesified Position Element Size
)
{
	List::POSITION pTarget = pos;
	if(pTarget != NULL)
	{
		if(pTarget->m_pData != NULL){
			pos = (POSITION) pTarget->m_pNext;
			*pLength = pTarget->m_dateLength;
			return (pTarget->m_pData);
		}
	}
	return NULL;
}

//----------------------------------------
//! GetPrevious
//! @return: Result(true:Success false:failure)
//----------------------------------------
bool List::GetPrevious(
	List::POSITION &pos, ///< [in/out] Next Node Pointer
	void* pBuffer, ///< [out] Spesified Position Element 
	int* pLength, ///< [out] Spesified Position Element Size
	int bufSize ///< [in] Prepared buffer sizes 
)
{
	List::POSITION pTarget = pos;
	if(pTarget != NULL)
	{
		if(pTarget->m_pData != NULL){
			if(bufSize < pTarget->m_dateLength)
			{
				return false;
			}
			pos = (POSITION) pTarget->m_pPrev;
			pafee_memcpy(pBuffer,pTarget->m_pData,pTarget->m_dateLength);
			*pLength = pTarget->m_dateLength;
			return true;
		}
	}
	return false;
}

//----------------------------------------
//! GetPrevious
//! @return: Result(true:Success false:failure)
//----------------------------------------
void* List::GetPrevious(
	List::POSITION &pos, ///< [in/out] Next Node Pointer
	int* pLength ///< [out] Spesified Position Element Size
)
{
	List::POSITION pTarget = pos;
	if(pTarget != NULL)
	{
		if(pTarget->m_pData != NULL){
			pos = (POSITION) pTarget->m_pPrev;
			*pLength = pTarget->m_dateLength;
			return pTarget->m_pData;
		}
	}
	return NULL;
}

//----------------------------------------
//! GetAt
//! @return: Result(true:Success false:failure)
//----------------------------------------
bool List::GetAt(
	List::POSITION pos,  ///< [in] Specified Position 
	void* pBuffer, ///< [out] Spesified Position Element 
	int* pLength, ///< [out] Spesified Position Element Size
	int bufSize ///< [in] Prepared buffer sizes 
)
{
	if(pos != NULL)
	{
		if(pos->m_pData != NULL){
			if(bufSize < pos->m_dateLength)
			{
				return false;
			}
			pafee_memcpy(pBuffer,pos->m_pData,pos->m_dateLength);
			*pLength = pos->m_dateLength;
			return true;
		}
	}
	return false;
}

//----------------------------------------
//! GetAt
//! @return: Result(true:Success false:failure)
//----------------------------------------
void* List::GetAt(
	List::POSITION pos, ///< [in] Specified Position 
	int* pLength ///< [out] Spesified Position Element Size
)
{
	if(pos != NULL)
	{
		if(pos->m_pData != NULL){
			*pLength = pos->m_dateLength;
			return pos->m_pData;
		}
	}
	return NULL;
}

//----------------------------------------
//! SetAt
//! @return: Result(true:Success false:failure)
//----------------------------------------
bool List::SetAt(
	List::POSITION& pos, ///< [in] Specified Position   
	const void* pData,   ///< [in] Spesified Position Element 
	int length         ///< [in] Spesified Position Element Size
)
{
	if(pos != NULL)
	{
		if(pos->m_pData != NULL){
			if(pos->m_bufferSize < length)
			{	
				if(pos->m_continues == false)
				{
					Deallocate((char* )pos->m_pData);
				}
				pos->m_pData = Allocate(length);
				pos->m_continues = false;
				pos->m_bufferSize = length;
			}
			pafee_memcpy(pos->m_pData,pData,length);
			pos->m_dateLength = length;
			return true;
		}
	}
	return false;
}

//----------------------------------------
//! RemoveFront
//! @return: void
//----------------------------------------
void List::RemoveFront()
{
	List::POSITION p;
	if(m_pStart == NULL)
	{
		return;
	}
	p = m_pStart;
	m_pStart = m_pStart->m_pNext;
	if(m_pStart != NULL)
	{
		m_pStart->m_pPrev = NULL;
	}
	else
	{
		m_pEnd = NULL;
	}
	DestroyBox(p);
	m_elementNum--;
	//
	STATISTICS_INC(m_Statistics.deleteCount);
	STATISTICS_DEC(m_Statistics.presentElementCount);
	//
}

//----------------------------------------
//! RemoveBack
//! @return: void
//----------------------------------------
void List::RemoveBack()
{
	List::POSITION p;
	if(m_pEnd == NULL)
	{
		return;
	}
	p = m_pEnd;
	m_pEnd = m_pEnd->m_pPrev;
	if(m_pEnd != NULL)
	{
		m_pEnd->m_pNext = NULL;
	}
	else
	{
		m_pStart = NULL;
	}
	DestroyBox(p);
	m_elementNum--;
	//
	STATISTICS_INC(m_Statistics.deleteCount);
	STATISTICS_DEC(m_Statistics.presentElementCount);
	//
}

//----------------------------------------
//! RemoveAt
//! @return: void
//----------------------------------------
void List::RemoveAt(
	List::POSITION pos ///< [in]  Specified Position
)
{
	if(pos->m_pPrev != NULL)
	{
		pos->m_pPrev->m_pNext = pos->m_pNext;
	}
	if(pos->m_pNext != NULL)
	{
		pos->m_pNext->m_pPrev = pos->m_pPrev;
	}
	if(pos == m_pStart)
	{
		m_pStart = pos->m_pNext;
	}
	if(pos == m_pEnd)
	{
		m_pEnd = pos->m_pPrev;;
	}
	DestroyBox(pos);
	m_elementNum--;
	if(m_elementNum == 0)
	{
		m_pStart = NULL;
		m_pEnd = NULL;
	}
	//
	STATISTICS_INC(m_Statistics.deleteCount);
	STATISTICS_DEC(m_Statistics.presentElementCount);
	//
	return;
}

//----------------------------------------
//! RemoveAll
//! @return: void
//----------------------------------------
void List::RemoveAll()
{
	while(m_pStart != NULL)
	{
		RemoveFront();
	}
}

//----------------------------------------
//! CreateBox
//! @return: Result(true:Success false:failure)
//----------------------------------------
List::POSITION List::CreateBox(
	const void* pData, ///< [in] Write Thing 
	int length       ///< [in] Write Size
)
{
	List::POSITION pCreate;
	pCreate = (List::POSITION)Allocate(sizeof(NODE_BOX) + length);
	if ( pCreate == NULL )
	{
		return NULL;
	}
	pCreate->m_pPrev = NULL;
	pCreate->m_pNext = NULL;
	pCreate->m_continues = true;
	pCreate->m_bufferSize = length;
	pCreate->m_pData = (char* )pCreate + sizeof(NODE_BOX);
	if(pData == NULL)
	{
		pCreate->m_dateLength = 0;
	}
	else
	{
		pCreate->m_dateLength = length;
		pafee_memcpy(pCreate->m_pData,pData,length);
	}
	return pCreate;
}

//----------------------------------------
//! DestroyBox
//! @return:    void
//----------------------------------------
void List::DestroyBox(
	List::POSITION pos ///< [in] Delete Position
)
{
	if((pos->m_continues == false) && (pos->m_pData != NULL))
	{
		Deallocate( (char* )pos->m_pData);
	}
	Deallocate( pos );
}

#ifdef PAFEE_STATISTICS_ENABLE
//----------------------------------------
//! GetStatistics
//! @return: void
//----------------------------------------
void List::GetStatistics(
	List::STATISTICS* pStatistics ///< [out] statistics Pointer
)
{
	List::POSITION p = GetFrontPosition();
	int dataLen;
	if(p == NULL)
	{
		m_Statistics.minElementSize = 0;
		m_Statistics.maxElementSize = 0;
	}
	else
	{
		for(p=m_pStart; p!=NULL; p=p->m_pNext)
		{
			dataLen = p->m_dateLength;
			if(p == m_pStart)
			{
				m_Statistics.minElementSize = dataLen;
				m_Statistics.maxElementSize = dataLen;

			}
			//min
			if(dataLen < m_Statistics.minElementSize)
			{
				m_Statistics.minElementSize = dataLen;
			}
			//max
			if(m_Statistics.maxElementSize < dataLen)
			{
				m_Statistics.maxElementSize = dataLen;
			}
		}
	}
	pafee_memcpy ( pStatistics, &m_Statistics, sizeof (List::STATISTICS) );
}
#endif //PAFEE_STATISTICS_ENABLE

#ifdef PAFEE_UNITTEST
//----------------------------------------------------------------------------
//!	Unit test
//! @return: void
//----------------------------------------------------------------------------
void List::UnitTest(void)
{
	{
		const int TEST_LIST_RECYCLE_COUNT = 10;
		int pData[TEST_LIST_RECYCLE_COUNT*2];
		int pOutput[TEST_LIST_RECYCLE_COUNT*2];
		int i;
		int size = sizeof(int);	
		int bufferSize = sizeof(int);///pOutput
		int length;
		int* p;
		//STATISTICS
		int addCount=0;
		int deleteCount=0;
		int maxElementCount=0;
		int minElementSize=0;
		int maxElementSize=0;
		int presentElementCount=0;
#ifdef PAFEE_STATISTICS_ENABLE
		List::STATISTICS statisticsRtn;
#endif
		List list;
		// initialize test data
		for(i=0; i<TEST_LIST_RECYCLE_COUNT*2; i++)
		{
			pData[i] = i;
		}
		//
		// test Initialize
		//
		///	assert(list.Initialize(TEST_LIST_RECYCLE_COUNT));
		//
		// no copy version test 
		//
		{	
			int count = 0;
			List::POSITION pos;
			List::POSITION padr;//void* padr;
			// test GetCount/IsEmpty
			assert(list.GetCount() == 0);	
			assert(list.IsEmpty());
			// Add test using GetWriteBuffer/AddFront/SetAt
			for(i = 0; i < TEST_LIST_RECYCLE_COUNT; i++)
			{
				p = (int* )list.GetWriteBuffer(size);
				assert(p!= NULL);
				*p = pData[i];
				assert(list.AddFront(size));
				pos = list.GetFrontPosition();
				assert(list.SetAt(pos, &pData[i], size));
				count++;
				assert(list.GetCount() == count);
				addCount++;
				presentElementCount++;
			}
			maxElementCount = TEST_LIST_RECYCLE_COUNT;
			minElementSize = size;
			maxElementSize = size;
			assert(!list.IsEmpty());
#ifdef PAFEE_STATISTICS_ENABLE
			list.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.deleteCount == deleteCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			// Find test using GetBackPosition/GetPrevious/GetPreviousLength
			padr = list.GetBackPosition();
			for(i = 0; i < TEST_LIST_RECYCLE_COUNT; i++)
			{
				length = list.GetPreviousLength(padr);///
				if(padr == list.GetFrontPosition())
				{
					assert(length == 0);
					break;
				}
				pOutput[i] = *((int*)list.GetPrevious(padr, &size));
				assert(pOutput[i] == i);
				assert(length == size);
			}

#ifdef PAFEE_STATISTICS_ENABLE
			list.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.deleteCount == deleteCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			// test RemoveAll
			list.RemoveAll();
			assert(list.GetCount() == 0);
			assert(list.IsEmpty());
			deleteCount += presentElementCount;
			minElementSize = 0;
			maxElementSize = 0;
			presentElementCount = 0;
#ifdef PAFEE_STATISTICS_ENABLE
			list.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.deleteCount == deleteCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			// Add test using GetWriteBuffer/AddBack/SetAt
			count = 0;
			for(i = 0; i < TEST_LIST_RECYCLE_COUNT; i++)
			{
				p = (int* )list.GetWriteBuffer(size);
				assert(p!=NULL);
				*p = pData[i];
				assert(list.AddBack(size));
				pos = list.GetBackPosition();
				assert(list.SetAt(pos, &pData[i], size));
				count++;
				assert(list.GetCount() == count);
				addCount++;
				presentElementCount++;
			}
			minElementSize = size;
			maxElementSize = size;
			assert(!list.IsEmpty());
#ifdef PAFEE_STATISTICS_ENABLE
			list.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.deleteCount == deleteCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			// Find test using GetFrontPosition/GetNext/GetNextPosition
			padr = list.GetFrontPosition();
			for(i = 0; i < TEST_LIST_RECYCLE_COUNT; i++)
			{
				length = list.GetNextLength(padr);
				if(padr == list.GetBackPosition())
				{
					assert(length == 0);
					break;
				}
				pOutput[i] = *((int*)list.GetNext(padr, &size));
				assert(pOutput[i] == i);
				assert(length == size);
			}
			int front, back, getAtValue;
			// test GetFront
			front = *((int*)list.GetFront(&size));
			assert(front == 0);
			// test GetBack
			back = *((int*)list.GetBack(&size));
			assert(back == 9);
			// test GetAt
			padr = list.GetFrontPosition();
			getAtValue = *((int*)list.GetAt(padr, &size));
			assert(getAtValue == 0);
#ifdef PAFEE_STATISTICS_ENABLE
			list.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.deleteCount == deleteCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			// test RemoveFront
			list.RemoveFront();
			front = *((int*)list.GetFront(&size));
			assert(front == 1);
			deleteCount++;
			presentElementCount--;
#ifdef PAFEE_STATISTICS_ENABLE
			list.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.deleteCount == deleteCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			// test RemoveBack
			list.RemoveBack();
			back = *((int*)list.GetBack(&size));
			assert(back == 8);
			deleteCount++;
			presentElementCount--;
#ifdef PAFEE_STATISTICS_ENABLE
			list.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.deleteCount == deleteCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			// test RemoveAt
			padr = list.GetFrontPosition();
			list.RemoveAt(padr);
			deleteCount++;
			presentElementCount--;
#ifdef PAFEE_STATISTICS_ENABLE
			list.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.deleteCount == deleteCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			padr = list.GetFrontPosition();
			getAtValue = *((int*)list.GetAt(padr, &size));
			assert(getAtValue == 2);
			// test RemoveAll
			list.RemoveAll();
			assert(list.GetCount() == 0);	
			assert(list.IsEmpty());
			deleteCount += presentElementCount;
			minElementSize = 0;
			maxElementSize = 0;
			presentElementCount = 0;
#ifdef PAFEE_STATISTICS_ENABLE
			list.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.deleteCount == deleteCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
		}
		//
		// test 
		//
		{
			int count = 0;
			List::POSITION padr;//void* padr;
			// test GetCount/IsEmpty
			assert(list.GetCount() == 0);	
			assert(list.IsEmpty());
			// Add test using GetFrontLength/GetBackLength/GetLengthAt
			for(i = 0; i < TEST_LIST_RECYCLE_COUNT; i++)
			{
				assert(list.AddBack(&pData[i], size));
				addCount++;
				presentElementCount++;
			}
			minElementSize = size;
			maxElementSize = size;
#ifdef PAFEE_STATISTICS_ENABLE
			list.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.deleteCount == deleteCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			assert(list.GetFrontLength() == size);
			assert(list.GetBackLength() == size);
			padr=list.GetFrontPosition();
			assert(list.GetNext(padr, &pOutput[i], &size, bufferSize));
			assert(list.GetNext(padr, &pOutput[i], &size, bufferSize));
			assert(list.GetLengthAt(padr) == size);
			list.RemoveAll();
			deleteCount += presentElementCount;
			minElementSize = 0;
			maxElementSize = 0;
			presentElementCount = 0;
#ifdef PAFEE_STATISTICS_ENABLE
			list.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.deleteCount == deleteCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			// Add test using AddFront
			for(i = 0; i < TEST_LIST_RECYCLE_COUNT; i++)
			{
				assert(list.AddFront(&pData[i], size));
				count++;
				assert(list.GetCount() == count);
				addCount++;
				presentElementCount++;
			}
			minElementSize = size;
			maxElementSize = size;
			assert(!list.IsEmpty());
#ifdef PAFEE_STATISTICS_ENABLE
			list.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.deleteCount == deleteCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			// Find test using GetBackPosition/GetPrevious/GetPreviousLength
			padr = list.GetBackPosition();
			for(i = 0; i < TEST_LIST_RECYCLE_COUNT; i++)
			{
				length = list.GetPreviousLength(padr);////GetPreviousLength
				if(padr == list.GetFrontPosition())
				{
					assert(length == 0);
					break;
				}
				assert(list.GetPrevious(padr, &pOutput[i], &size, bufferSize));
				assert(pOutput[i] == i);
				assert(length == size);
			}
#ifdef PAFEE_STATISTICS_ENABLE
			list.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.deleteCount == deleteCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			// test RemoveAll
			list.RemoveAll();
			assert(list.GetCount() == 0);
			assert(list.IsEmpty());
			deleteCount += presentElementCount;
			minElementSize = 0;
			maxElementSize = 0;
			presentElementCount = 0;
#ifdef PAFEE_STATISTICS_ENABLE
			list.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.deleteCount == deleteCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			// Add test using AddBack
			count = 0;
			for(i = 0; i < TEST_LIST_RECYCLE_COUNT; i++)
			{
				assert(list.AddBack(&pData[i], size));
				count++;
				assert(list.GetCount() == count);
				addCount++;
				presentElementCount++;
			}
			minElementSize = size;
			maxElementSize = size;
			assert(!list.IsEmpty());
#ifdef PAFEE_STATISTICS_ENABLE
			list.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.deleteCount == deleteCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			// Find test using GetFrontPosition/GetNext/GetNextLength
			padr = list.GetFrontPosition();
			for(i = 0; i < TEST_LIST_RECYCLE_COUNT; i++)
			{
				length = list.GetNextLength(padr);
				if(padr == list.GetBackPosition())
				{
					assert(length == 0);
					break;
				}
				assert(list.GetNext(padr, &pOutput[i], &size, bufferSize));
				assert(pOutput[i] == i);
				assert(length == size);
			}
			int front, back, getAtValue;
			// test GetFront
			list.GetFront(&front, &size, bufferSize);
			assert(front == 0);
			// test GetBack
			list.GetBack(&back, &size, bufferSize);
			assert(back == 9);
			// test GetAt
			padr = list.GetFrontPosition();
			list.GetAt(padr, &getAtValue, &size, bufferSize);
			assert(getAtValue == 0);
			// test SetAt
			padr = list.GetFrontPosition();
			assert(list.SetAt(padr, &pData[i], size));
#ifdef PAFEE_STATISTICS_ENABLE
			list.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.deleteCount == deleteCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			// test RemoveFront
			list.RemoveFront();
			list.GetFront(&front, &size, bufferSize);
			assert(front == 1);
			deleteCount++;
			presentElementCount--;
#ifdef PAFEE_STATISTICS_ENABLE
			list.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.deleteCount == deleteCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			// test RemoveBack
			list.RemoveBack();
			list.GetBack(&back, &size, bufferSize);
			assert(back == 8);
			deleteCount++;
			presentElementCount--;
#ifdef PAFEE_STATISTICS_ENABLE
			list.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.deleteCount == deleteCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			// test RemoveAt
			padr = list.GetFrontPosition();
			list.RemoveAt(padr);
			padr = list.GetFrontPosition();
			list.GetAt(padr, &getAtValue, &size, bufferSize);
			assert(getAtValue == 2);
			deleteCount++;
			presentElementCount--;
#ifdef PAFEE_STATISTICS_ENABLE
			list.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.deleteCount == deleteCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif

			//RemoveAll
			list.RemoveAll();
			assert(list.GetCount() == 0);
			assert(list.IsEmpty());
			deleteCount += presentElementCount;
			minElementSize = 0;
			maxElementSize = 0;
			presentElementCount = 0;
#ifdef PAFEE_STATISTICS_ENABLE
			list.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.deleteCount == deleteCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
		}
	}

////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////

	{
		const int TEST_LIST_RECYCLE_COUNT = 10;
		List list0c;
		int pData[TEST_LIST_RECYCLE_COUNT*2];
		int pOutput[TEST_LIST_RECYCLE_COUNT*2];
		int i;
		int size = sizeof(int);	
		int bufferSize = sizeof(int);///pOutput
		int length;
		int* p;
		//STATISTICS
		int addCount=0;
		int deleteCount=0;
		int maxElementCount=0;
		int minElementSize=0;
		int maxElementSize=0;
		int presentElementCount=0;
#ifdef PAFEE_STATISTICS_ENABLE
		List::STATISTICS statisticsRtn;
#endif
		// initialize test data
		for(i=0; i<TEST_LIST_RECYCLE_COUNT*2; i++)
		{
			pData[i] = i;
		}
		//
		// test Initialize
		//
		///	assert(list.Initialize(TEST_LIST_RECYCLE_COUNT));
		//
		// no copy version test 
		//
		{	
			int count = 0;
			List::POSITION pos;
			List::POSITION padr;//void* padr;
			// test GetCount/IsEmpty
			assert(list0c.GetCount() == 0);	
			assert(list0c.IsEmpty());
			// Add test using GetWriteBuffer/AddFront/SetAt
			for(i = 0; i < TEST_LIST_RECYCLE_COUNT; i++)
			{
				p = (int* )list0c.GetWriteBuffer(size);
				assert(p!= NULL);
				*p = pData[i];
				assert(list0c.AddFront(size));
				pos = list0c.GetFrontPosition();
				assert(list0c.SetAt(pos, &pData[i], size));
				count++;
				assert(list0c.GetCount() == count);
				addCount++;
				presentElementCount++;
			}
			maxElementCount = TEST_LIST_RECYCLE_COUNT;
			minElementSize = size;
			maxElementSize = size;
			assert(!list0c.IsEmpty());
#ifdef PAFEE_STATISTICS_ENABLE
			list0c.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.deleteCount == deleteCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			List list =list0c;
			// Find test using GetBackPosition/GetPrevious/GetPreviousLength
			padr = list.GetBackPosition();
			for(i = 0; i < TEST_LIST_RECYCLE_COUNT; i++)
			{
				length = list.GetPreviousLength(padr);///
				if(padr == list.GetFrontPosition())
				{
					assert(length == 0);
					break;
				}
				pOutput[i] = *((int*)list.GetPrevious(padr, &size));
				assert(pOutput[i] == i);
				assert(length == size);
			}
#ifdef PAFEE_STATISTICS_ENABLE
			list.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.deleteCount == deleteCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			// test RemoveAll
			list.RemoveAll();
			assert(list.GetCount() == 0);
			assert(list.IsEmpty());
			deleteCount += presentElementCount;
			minElementSize = 0;
			maxElementSize = 0;
			presentElementCount = 0;
#ifdef PAFEE_STATISTICS_ENABLE
			list.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.deleteCount == deleteCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			// Add test using GetWriteBuffer/AddBack/SetAt
			count = 0;
			for(i = 0; i < TEST_LIST_RECYCLE_COUNT; i++)
			{
				p = (int* )list.GetWriteBuffer(size);
				assert(p!=NULL);
				*p = pData[i];
				assert(list.AddBack(size));
				pos = list.GetBackPosition();
				assert(list.SetAt(pos, &pData[i], size));
				count++;
				assert(list.GetCount() == count);
				addCount++;
				presentElementCount++;
			}
			minElementSize = size;
			maxElementSize = size;
			assert(!list.IsEmpty());
#ifdef PAFEE_STATISTICS_ENABLE
			list.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.deleteCount == deleteCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			// Find test using GetFrontPosition/GetNext/GetNextPosition
			padr = list.GetFrontPosition();
			for(i = 0; i < TEST_LIST_RECYCLE_COUNT; i++)
			{
				length = list.GetNextLength(padr);
				if(padr == list.GetBackPosition())
				{
					assert(length == 0);
					break;
				}
				pOutput[i] = *((int*)list.GetNext(padr, &size));
				assert(pOutput[i] == i);
				assert(length == size);
			}

			int front, back, getAtValue;
			// test GetFront
			front = *((int*)list.GetFront(&size));
			assert(front == 0);
			
			// test GetBack
			back = *((int*)list.GetBack(&size));
			assert(back == 9);

			// test GetAt
			padr = list.GetFrontPosition();
			getAtValue = *((int*)list.GetAt(padr, &size));
			assert(getAtValue == 0);
#ifdef PAFEE_STATISTICS_ENABLE
			list.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.deleteCount == deleteCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			// test RemoveFront
			list.RemoveFront();
			front = *((int*)list.GetFront(&size));
			assert(front == 1);
			deleteCount++;
			presentElementCount--;
#ifdef PAFEE_STATISTICS_ENABLE
			list.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.deleteCount == deleteCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			// test RemoveBack
			list.RemoveBack();
			back = *((int*)list.GetBack(&size));
			assert(back == 8);
			deleteCount++;
			presentElementCount--;
#ifdef PAFEE_STATISTICS_ENABLE
			list.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.deleteCount == deleteCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			// test RemoveAt
			padr = list.GetFrontPosition();
			list.RemoveAt(padr);
			deleteCount++;
			presentElementCount--;
#ifdef PAFEE_STATISTICS_ENABLE
			list.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.deleteCount == deleteCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			padr = list.GetFrontPosition();
			getAtValue = *((int*)list.GetAt(padr, &size));
			assert(getAtValue == 2);
			// test RemoveAll
			list.RemoveAll();
			assert(list.GetCount() == 0);	
			assert(list.IsEmpty());
			deleteCount += presentElementCount;
			minElementSize = 0;
			maxElementSize = 0;
			presentElementCount = 0;
#ifdef PAFEE_STATISTICS_ENABLE
			list.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.deleteCount == deleteCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
		}
		//
		// test 
		//
		{
			int addCount=0;
			int deleteCount=0;
			int maxElementCount=0;
			int minElementSize=0;
			int maxElementSize=0;
			int presentElementCount=0;
			List list0c;
			int count = 0;
			List::POSITION padr;//void* padr;
			// test GetCount/IsEmpty
			assert(list0c.GetCount() == 0);	
			assert(list0c.IsEmpty());
			// Add test using GetFrontLength/GetBackLength/GetLengthAt
			for(i = 0; i < TEST_LIST_RECYCLE_COUNT; i++)
			{
				assert(list0c.AddBack(&pData[i], size));
				addCount++;
				presentElementCount++;
				maxElementCount++;
			}
			minElementSize = size;
			maxElementSize = size;
#ifdef PAFEE_STATISTICS_ENABLE
			list0c.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.deleteCount == deleteCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			assert(list0c.GetFrontLength() == size);
			assert(list0c.GetBackLength() == size);
			padr=list0c.GetFrontPosition();
			assert(list0c.GetNext(padr, &pOutput[i], &size, bufferSize));
			assert(list0c.GetNext(padr, &pOutput[i], &size, bufferSize));
			assert(list0c.GetLengthAt(padr) == size);
			list0c.RemoveAll();
			deleteCount += presentElementCount;
			minElementSize = 0;
			maxElementSize = 0;
			presentElementCount = 0;
#ifdef PAFEE_STATISTICS_ENABLE
			list0c.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.deleteCount == deleteCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			// Add test using AddFront
			for(i = 0; i < TEST_LIST_RECYCLE_COUNT; i++)
			{
				assert(list0c.AddFront(&pData[i], size));
				count++;
				assert(list0c.GetCount() == count);
				addCount++;
				presentElementCount++;
			}
			minElementSize = size;
			maxElementSize = size;
			assert(!list0c.IsEmpty());
#ifdef PAFEE_STATISTICS_ENABLE
			list0c.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.deleteCount == deleteCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			//List list =list0c;
			List list;
			list =list0c;
			// Find test using GetBackPosition/GetPrevious/GetPreviousLength
			padr = list.GetBackPosition();
			for(i = 0; i < TEST_LIST_RECYCLE_COUNT; i++)
			{
				length = list.GetPreviousLength(padr);////GetPreviousLength
				if(padr == list.GetFrontPosition())
				{
					assert(length == 0);
					break;
				}
				assert(list.GetPrevious(padr, &pOutput[i], &size, bufferSize));
				assert(pOutput[i] == i);
				assert(length == size);
			}
#ifdef PAFEE_STATISTICS_ENABLE
			list.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.deleteCount == deleteCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			// test RemoveAll
			list.RemoveAll();
			assert(list.GetCount() == 0);
			assert(list.IsEmpty());
			deleteCount += presentElementCount;
			minElementSize = 0;
			maxElementSize = 0;
			presentElementCount = 0;
#ifdef PAFEE_STATISTICS_ENABLE
			list.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.deleteCount == deleteCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			// Add test using AddBack
			count = 0;
			for(i = 0; i < TEST_LIST_RECYCLE_COUNT; i++)
			{
				assert(list.AddBack(&pData[i], size));
				count++;
				assert(list.GetCount() == count);
				addCount++;
				presentElementCount++;
			}
			minElementSize = size;
			maxElementSize = size;
			assert(!list.IsEmpty());
#ifdef PAFEE_STATISTICS_ENABLE
			list.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.deleteCount == deleteCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			// Find test using GetFrontPosition/GetNext/GetNextLength
			padr = list.GetFrontPosition();
			for(i = 0; i < TEST_LIST_RECYCLE_COUNT; i++)
			{
				length = list.GetNextLength(padr);
				if(padr == list.GetBackPosition())
				{
					assert(length == 0);
					break;
				}
				assert(list.GetNext(padr, &pOutput[i], &size, bufferSize));
				assert(pOutput[i] == i);
				assert(length == size);
			}
			int front, back, getAtValue;
			// test GetFront
			list.GetFront(&front, &size, bufferSize);
			assert(front == 0);
			// test GetBack
			list.GetBack(&back, &size, bufferSize);
			assert(back == 9);
			// test GetAt
			padr = list.GetFrontPosition();
			list.GetAt(padr, &getAtValue, &size, bufferSize);
			assert(getAtValue == 0);
			// test SetAt
			padr = list.GetFrontPosition();
			assert(list.SetAt(padr, &pData[i], size));
#ifdef PAFEE_STATISTICS_ENABLE
			list.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.deleteCount == deleteCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			// test RemoveFront
			list.RemoveFront();
			list.GetFront(&front, &size, bufferSize);
			assert(front == 1);
			deleteCount++;
			presentElementCount--;
#ifdef PAFEE_STATISTICS_ENABLE
			list.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.deleteCount == deleteCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			// test RemoveBack
			list.RemoveBack();
			list.GetBack(&back, &size, bufferSize);
			assert(back == 8);
			deleteCount++;
			presentElementCount--;
#ifdef PAFEE_STATISTICS_ENABLE
			list.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.deleteCount == deleteCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			// test RemoveAt
			padr = list.GetFrontPosition();
			list.RemoveAt(padr);
			padr = list.GetFrontPosition();
			list.GetAt(padr, &getAtValue, &size, bufferSize);
			assert(getAtValue == 2);
			deleteCount++;
			presentElementCount--;
#ifdef PAFEE_STATISTICS_ENABLE
			list.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.deleteCount == deleteCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
			//RemoveAll
			list.RemoveAll();
			assert(list.GetCount() == 0);
			assert(list.IsEmpty());
			deleteCount += presentElementCount;
			minElementSize = 0;
			maxElementSize = 0;
			presentElementCount = 0;
#ifdef PAFEE_STATISTICS_ENABLE
			list.GetStatistics(&statisticsRtn);
			assert(statisticsRtn.addCount == addCount);
			assert(statisticsRtn.deleteCount == deleteCount);
			assert(statisticsRtn.maxElementCount == maxElementCount);
			assert(statisticsRtn.minElementSize == minElementSize);
			assert(statisticsRtn.maxElementSize == maxElementSize);
			assert(statisticsRtn.presentElementCount == presentElementCount);
#endif
		}
	}
}
#endif

}
