//****************************************************
//
//    project   Pafee
//!   @file:    PafeeMultiMap.h
//!   @author:  M.Tanaka
//!   @brief:   MultiMap Hedder File
//
//*****************************************************

//
// P A F E E   - Pragmatic Application Framework for Embedded Environment- 
//
//
// Copyright (c) 2009 KN Information Systems Corporation,
//                    Hirocon Corporation,
//                    Technocrats Corporation
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//

#include <new>
#include <limits.h>
#include "PafeeList.h"
#include "pafee_string.h"
#include "PafeeCommon.h"
#include "PafeeObject.h"
#include "PafeeAllocator.h"

#ifdef PAFEE_UNITTEST
#include <assert.h>
#endif //PAFEE_UNITTEST

#ifndef PAFEE_MULTI_MAP
#define PAFEE_MULTI_MAP

namespace Pafee
{

//! MultiMap
	class MultiMap : public Object
	{

#ifdef PAFEE_STATISTICS_ENABLE
public:
	typedef struct
	{
		int addCount;
		int acqCount;
		int deleteCount;
		int maxElementCount;
		int collisionCount;
		int minListCount;
		int maxListCount;
		int presentElementCount;
	} STATISTICS;
	void GetStatistics(MultiMap::STATISTICS* pStatistics);
private:
	STATISTICS m_Statistics;
#endif // PAFEE_STATISTICS_ENABLE

	public:
		struct POSITION
		{
			List::POSITION pos;
			int m_positionIndex;
			const char* m_pCharkey;
			int m_numberKey;
		};
	public:
		MultiMap(int hashSize=0, Allocator* pAllocator=NULL);
		MultiMap(const MultiMap& mmpc, Allocator* pAllocator=NULL);
		~ MultiMap(); 
		bool Initialize(int hashSize, Allocator* pAllocator=NULL);
		MultiMap& operator = (const MultiMap& mmpc);
		int GetCount(void)
		{
			return m_size;
		}
		bool IsEmpty(void)
		{
			return !m_size;
		}
		void* GetWriteBuffer(const char* szKey, int length);
		void* GetWriteBuffer(const int key, int length);
		bool Set(const char* szKey, void* pData, int length);
		bool Set(int key, void* pData, int length);
		bool Set (int length); //No copy version
		MultiMap::POSITION GetStartPosition(void);
		MultiMap::POSITION GetStartPosition(const char* szKey);
		MultiMap::POSITION GetStartPosition(int key);
		bool GetNext(MultiMap::POSITION& pos, void* pBuffer, int* pLength, int bufSize=INT_MAX);
		void* GetNext(MultiMap::POSITION& pos, int* pLength);//No copy version
		void Remove(const char* szKey);
		void Remove(int key);
		void RemoveAll(void);

	private:
		int m_indexCount;
		List* m_pHashTable;
		int m_hashSize;
		int m_size;

		unsigned int Hash( const char *pHashKey )
		{
			unsigned int ikey=0;
			for (; *(pHashKey+3)!='\0';)
			{
				ikey ^= *(unsigned int*)pHashKey;
				(unsigned int*)pHashKey++;
			}
			return ikey%m_hashSize;
		}

		unsigned int Hash( int hashKey )
		{
			unsigned int hHight  = hashKey ^ 0xFFFF0000;
			unsigned int hLow    = hashKey ^ 0x0000FFFF;
			unsigned int hLow2   = (hHight >> 16) ^ 0x0000FFFF;
			unsigned int hHight2 = (hLow << 16)   ^ 0xFFFF0000;	
			unsigned int uhashKey = hHight2 | hLow2;
			unsigned int rhashKey = uhashKey ^ hashKey;	
			return (rhashKey%m_hashSize) & 0x7FFFFFFF;
		}


		int GetBufferSize(const int key, int length);
		int GetBufferSize(const char* szKey, int length);
		char* SetBuffer(char* pBuffer, const int key, int length);
		char* SetBuffer(char* pBuffer, const char* szKey, int length);
		int GetKeyType(const char* pElement);
		char* GetKey( char* pData, int* pKeyNum );
		char* GetData( char* pIn, int* pLength );
		int GetData( char* pOut, const char* pIn );
		int GetDataLength( const char* pIn );
		char* SetData(const int key, const char* pData, int length, int* pBufferSize);
		char* SetData(const char* szKey, const char* pData, int length, int* pBufferSize);
		//
		void GetUsedCharSize( const int key, int length, char* usedCharNum1, char* usedCharNum2);
		void GetUsedCharSize( const char* szKey, int length, char* usedCharNum, char* usedCharNum1, unsigned int* usedCharNum2);
		void GetDataSub( const char* pBuffer, int* pLengthToData, int* pDataLength );
		int SetDataSub( char* pBuffer, const int key, int length, char usedCharNum1, char usedCharNum2 );
		int SetDataSub( char* pBuffer, const char* szKey, int length, char usedCharNum, char usedCharNum1, unsigned int usedCharNum2 );
		//
		char* BufferAllocate( int size )
		{
			char* pNewBuffer;
			pNewBuffer = (char* )Allocate(AligmentSize(size));
			return pNewBuffer;
		}
		int m_writeIndex;
		char* m_pWriteBuffer;
		int m_writeBufferSize;

	private:
		inline int AligmentSize(int size) {return ((((size)-1)|(((size)-1)^3))+1);};

#ifdef PAFEE_UNITTEST
	public: // unit test function
		static void UnitTest(void);
#endif // PAFEE_UNITTEST
	};
}
#endif
