//********************************************************************
//		project 	PAFEE
//
//! 	 @file: 	PafeeFastFixedAllocator.h
//! 	 @author:	Takatsura Miyawaki
//! 	 @brief:	fast fixed memory allocator
//
//********************************************************************

//
// P A F E E   - Pragmatic Application Framework for Embedded Environment- 
//
//
// Copyright (c) 2009 KN Information Sustems Corporation,
//                    Hirocon Corporation,
//                    Technocrats Corporation
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//

#ifndef _FAST_FIXED_ALLOCATOR_H_INCLUDED_
#define _FAST_FIXED_ALLOCATOR_H_INCLUDED_

#include <stddef.h>
#include "PafeeCommon.h"
#include "pafee_string.h"
#include "PafeeAllocator.h"

namespace Pafee
{
	//!  FastFixedAllocator class
	class FastFixedAllocator : public Allocator
	{
	public:

		//! Constructor
		FastFixedAllocator();

		//! Destructor
		virtual ~FastFixedAllocator();

		//! Initialize
		bool Initialize(unsigned long blockSize, unsigned long blockCount, Allocator* pBackendAllocator=NULL);

		//! Allocate memory
		virtual void* Allocate(unsigned long size);

		//! Deallocate memory
		virtual void Deallocate(void* ptr);

		//! Check given address is in own address range
		virtual bool IsOwnAddress(void* ptr);


#ifdef PAFEE_STATISTICS_ENABLE
		//! statistics information
		typedef struct _STATISTICS
		{ 
			unsigned long allocateRequests;		///< count of allocate request
			unsigned long deallocateRequests;	///< count of deallocate request
			unsigned long failedAllocateRequests;///< count of allocation failure
			unsigned long sanityErrors;			///< count of sanity error
			unsigned long currentAllocations;	///< current allocated block count
			unsigned long peakAllocations;		///< peak allocated block count
			unsigned long minRequestSize;		///< minimum request size
			unsigned long maxRequestSize;		///< maximum request size
		} STATISTICS;

		//! Get statistics information
		void GetStatistics(FastFixedAllocator::STATISTICS* pStat);

		//! Get statistics
		virtual void GetStatistics(StreamOutput& stream);
#endif
#ifdef PAFEE_UNITTEST
		//! Unit test
		static void UnitTest(void);
#endif

	protected:
		static const unsigned long  SANITY_CODE;	///< block sanity check code
		static const unsigned long  SANITY_MASK;	///< mask for sanity code
		static const unsigned long  USEFLAG;		///< using flag
		static const unsigned long  OWNINDEX_MASK;	///< mask for own index

		//! block header
		typedef struct _BLK_HEADER
		{
			unsigned long	sanity_use_index;	///< sanity code, use flag and own index
			unsigned long	nextAvailableIndex;	///< next available index (### while this block is allocated, nextAvailableIndex field turns to be head of data buffer)
		} BLK_HEADER;
		
		unsigned long	m_blockSize;			///< block size
		unsigned long	m_blockCount;			///< block count
		unsigned long	m_pageSize;				///< page size;
		unsigned long	m_firstAvailableIndex;	///< index of first available block
		unsigned char*	m_pBlocks;				///< block buffers
		unsigned char*	m_pOriginalBuffer;		///< original buffer address
		Allocator* 		m_pBackendAllocator;	///< back-end allocator
		
#ifdef PAFEE_STATISTICS_ENABLE
		STATISTICS m_statistics;				///< statistics information
#endif


	private:
		FastFixedAllocator(const FastFixedAllocator& rhs);
		FastFixedAllocator& operator=(const FastFixedAllocator& rhs);
	};

};

#endif // _FAST_FIXED_ALLOCATOR_H_INCLUDED_
