//********************************************************************
//		project 	PAFEE
//
//! 	 @file: 	PafeeFixedAllocator.cpp
//! 	 @brief:	fixed memory allocation
//
//********************************************************************

//
// P A F E E   - Pragmatic Application Framework for Embedded Environment- 
//
//
// Copyright (c) 2009 KN Information Sustems Corporation,
//                    Hirocon Corporation,
//                    Technocrats Corporation
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//
// #### NOTE ####
//
// This FixedAllocator is using the allocation algorithm of Loki::SmallObjAllocator by Alexandrescu, Andrei.
//
//

// modification history
//-----------------------
//
// 2009.02.11 
//	- added license notice
//	- changed exception
//	- changed data type
//	- forbid copying
//	- changed to use static const instread of #define
//	- moved global struct into class scope
//	- moved global define into class scope
//	- changed to use Allocator instead of malloc
//
// 2009.01.26 
//	- created.

#include <stddef.h>


#include "PafeeCommon.h"
#include "pafee_string.h"
#include "PafeeFixedAllocator.h"
#include "PafeeStdAllocator.h"

using namespace Pafee;

const unsigned long  FixedAllocator::CHUNK_HEADER_SIZE = 8;			///< chunk header size
const unsigned long  FixedAllocator::PAGE_HEADER_SIZE = 16;			///< page header size
const unsigned short FixedAllocator::PAGE_SANITY_CODE = 0xabab;		///< page sanity check code
const unsigned long  FixedAllocator::CHUNK_SANITY_CODE = 0xcdcdcdcd;///< chunk sanity check code

//--------------------------------------------------------------------
//!  initialize memory management system
//!  @return: page information table or NULL (failure)
//--------------------------------------------------------------------
FixedAllocator::PAGE* FixedAllocator::IntializePage(
		unsigned long chunkSize, 	///< [in] chunk size in page
		unsigned long chunks 		///< [in] amount of chunks in page
		)
{
	CHUNK* chunk;
	PAGE* page;
	char* p;
	unsigned short i;

	if (chunkSize < sizeof(unsigned long))
	{
		return NULL;
	}
	if (chunks == 0)
	{
		return NULL;
	}

	unsigned long pageSize = PAGE_HEADER_SIZE + (chunkSize+CHUNK_HEADER_SIZE)*chunks;
	if(m_pBackendAllocator)
	{
		page = (PAGE*) m_pBackendAllocator->Allocate(pageSize);
	}
	else
	{
		page = (PAGE*) StdAllocator::Malloc(pageSize);
	}
	if (page == NULL)
	{
		return NULL;
	}

	page->m_SanityCode = PAGE_SANITY_CODE;
	page->m_pNext = NULL;
	page->m_pPrevious = NULL;
	page->m_pThis = this;
	page->m_FirstAvailableChunk = 0;

	p = page->m_Data;
	for(i=0, p=page->m_Data; i<chunks; p+=(chunkSize+CHUNK_HEADER_SIZE))
	{
		chunk = (CHUNK*)p;
		i++;
		chunk->m_SanityCode = CHUNK_SANITY_CODE;
		chunk->m_PageOffset = (unsigned long)(p - (char*)page);
		chunk->m_Data[0] = i;
	}

	return page;
}

//--------------------------------------------------------------------
//!  initialize memory management system
//!  @return: allocated memory pointer or NULL(failure)
//--------------------------------------------------------------------
void* FixedAllocator::AllocateChunk(
		FixedAllocator::PAGE* page,	///< [in] pointer for page information table
		unsigned long chunkSize, 	///< [in] chunk size in page
		unsigned long chunks 		///< [in] amount of chunks in page
		)
{
	CHUNK* chunk;

	//PAFEE_TRACE("AllocateChunk:%ld %ld\n,", chunkSize, chunks);
	if (chunkSize == 0 || chunks == 0)
	{
		return NULL;
	}

	if (page == NULL)
	{
		return NULL;
	}
	if (page->m_SanityCode != PAGE_SANITY_CODE)
	{
		return NULL;
	}
	if (page->m_FirstAvailableChunk >= chunks)
	{
		return NULL;
	}

	chunk = (CHUNK*)(page->m_Data + (page->m_FirstAvailableChunk * (chunkSize+CHUNK_HEADER_SIZE)));
	if (chunk->m_SanityCode != CHUNK_SANITY_CODE)
	{
		return NULL;
	}
	page->m_FirstAvailableChunk = (unsigned short)chunk->m_Data[0];

	return chunk->m_Data;
}

//--------------------------------------------------------------------
//!  initialize memory management system
//!  @return: result (1:success , 0:failure)
//--------------------------------------------------------------------
bool FixedAllocator::DeallocateChunk(
		FixedAllocator::PAGE* page, ///< [in] pointer for page information table
		void* ptr,					///< [in] memory pointer
		unsigned long chunkSize,	///< [in] chunk size in page
		unsigned long chunks 		///< [in] amount of chunks in page
		)
{
	CHUNK* chunk;

	PAFEE_ASSERT(page);
	PAFEE_ASSERT(chunkSize);
	PAFEE_ASSERT(chunks);
	
	if (((char*)ptr < (char*)(page->m_Data)) || ((char*)ptr >= (char*)(page->m_Data) + (chunkSize+CHUNK_HEADER_SIZE)*chunks))
	{
		return false;
	}
	//toRelease = ptr;
	chunk = (CHUNK*)((char*)ptr - CHUNK_HEADER_SIZE);
	if (((char*)chunk - (char*)page->m_Data) % (chunkSize+CHUNK_HEADER_SIZE) != 0)
	{
		return false;
	}

	chunk->m_Data[0] = page->m_FirstAvailableChunk;
	page->m_FirstAvailableChunk = (unsigned short)(((char*)chunk - (char*)page->m_Data) / (chunkSize + CHUNK_HEADER_SIZE));

	return true;
}


#ifdef PAFEE_STATISTICS_ENABLE
//--------------------------------------------------------------------
//!  link page to pool
//!  @return: void
//--------------------------------------------------------------------
void FixedAllocator::CalcStatistics(
		STATISTICS* pStat, 	///< [in/out] statistics table pointer
		unsigned long requestSize, 	///< [in] request memory size
		unsigned long defaultSize 	///< [in] default page size
		)
{
	unsigned long raito;

	raito = (requestSize * 100) / defaultSize;

	if (raito < 20)
	{
		(pStat->m_RequestSizeUnder20PerChunk)++;
	}
	else if (raito < 40)
	{
		(pStat->m_RequestSizeUnder40PerChunk)++;
	}
	else if (raito < 60)
	{
		(pStat->m_RequestSizeUnder60PerChunk)++;
	}
	else if (raito < 80)
	{
		(pStat->m_RequestSizeUnder80PerChunk)++;
	}
	else
	{
		(pStat->m_RequestSizeUnder100PerChunk)++;
	}

	return;
}
#endif

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// public members

//--------------------------------------------------------------------
//!  constructor
//--------------------------------------------------------------------
FixedAllocator::FixedAllocator()
{
	m_AllocPage = NULL;
	m_PageList = NULL; 
	m_ChunkSize = 0; 
	m_Chunks = 0;
	m_pBackendAllocator = NULL;
}

//--------------------------------------------------------------------
//!  destructor
//--------------------------------------------------------------------
FixedAllocator::~FixedAllocator()
{
	PAGE* pPage = m_PageList;
	while(pPage)
	{
		PAGE* pNext = pPage->m_pNext;;
		if(m_pBackendAllocator)
		{
			m_pBackendAllocator->Deallocate(pPage);
		}
		else
		{
			StdAllocator::Free(pPage);
		}
		pPage = pNext;
	}
}

//--------------------------------------------------------------------
//!  initialize allocator
//!  @return: void
//--------------------------------------------------------------------
bool FixedAllocator::Initialize(
		unsigned long chunkSize, 	///< [in] chunk size in page
		unsigned long chunkAmount, 	///< [in] amount of chunk in page (size*amount=page size)
		unsigned long initialPages,	///< [in] amount of page
		Allocator* pBackendAllocator///< [in] backend allocator
		)
{
	if ((chunkAmount == 0) ||  (chunkSize < sizeof(unsigned long)))
	{
		return false;
	}
	chunkSize = AlignUp(chunkSize);
	
	if (initialPages == 0)
	{
		initialPages = 1;
	}
	m_pBackendAllocator = pBackendAllocator;

	m_ChunkSize = chunkSize;
	m_Chunks = chunkAmount;
	m_PageList = NULL;

#ifdef PAFEE_STATISTICS_ENABLE
	pafee_memset((&(m_Statistics)), ('\0'), (sizeof(m_Statistics)));
	m_Statistics.m_RequestMemorySizeMin = 0xffffffff;
#endif

	PAGE* page = NULL;
	for (unsigned long i=0; i < initialPages; i++)
	{
		PAGE* next = IntializePage(m_ChunkSize, m_Chunks);
		if (next == NULL)
		{
			break;
		}
		if (page == NULL)
		{
			m_PageList = next;
		}
		else
		{
			page->m_pNext = next;
		}
		page = next;

		STATISTICS_INC(m_Statistics.m_GetHeap);
		STATISTICS_INC(m_Statistics.m_UsingPageSizeAll);
	}
	m_AllocPage = m_PageList;
	return true;
}

//--------------------------------------------------------------------
//!  memory allocation
//!  @return: allocated memory pointer (fixed size) or NULL(failure)
//--------------------------------------------------------------------
void* FixedAllocator::Allocate(
		unsigned long  size ///< [in] memory size (not used)
		)
{
	//PAFEE_TRACE("Allocate:%ld - ", size);
	if (size > m_ChunkSize)
	{
#ifdef PAFEE_EXCEPTION
		throw std::runtime_error("size over");
#endif
		return NULL;
	}

	STATISTICS_INC(m_Statistics.m_AllocateRequest);
	STATISTICS_MIN(m_Statistics.m_RequestMemorySizeMin, size);
	STATISTICS_MAX(m_Statistics.m_RequestMemorySizeMax, size);
#ifdef PAFEE_STATISTICS_ENABLE
	CalcStatistics(&(m_Statistics), size, m_ChunkSize);
#endif

	if (m_AllocPage == NULL)
	{
		m_AllocPage = m_PageList;
	}
	if (m_AllocPage->m_FirstAvailableChunk >= m_Chunks)
	{
		for (PAGE* page = m_PageList; page != NULL; page = page->m_pNext)
		{
			if (page->m_FirstAvailableChunk < m_Chunks)
			{
				m_AllocPage = page;
				break;
			}
			if (page->m_pNext == NULL)
			{
				page->m_pNext = IntializePage(m_ChunkSize, m_Chunks);
				if (page->m_pNext == NULL)
				{
					return NULL;
				}
				m_AllocPage = page->m_pNext;
				STATISTICS_INC(m_Statistics.m_GetHeap);
				STATISTICS_INC(m_Statistics.m_UsingPageSizeAll);
				break;
			}
		}
	}

	STATISTICS_ADD(m_Statistics.m_UsingMemorySizeAll, m_ChunkSize);

	void* ptr = AllocateChunk(m_AllocPage, m_ChunkSize, m_Chunks);
#ifdef PAFEE_EXCEPTION
	if(ptr == NULL)
	{
		throw std::bad_alloc();
	}
#endif
	//PAFEE_TRACE("%p\n", ptr);
	return ptr;
}

//--------------------------------------------------------------------
//!  memory free (deallocation)
//!  @return: void
//--------------------------------------------------------------------
void FixedAllocator::Deallocate(
		void* ptr ///< [in] allocated memory pointer
		)
{
	//PAFEE_TRACE("Deallocate:%p \n", ptr);
	if (ptr == NULL) // normal
	{
		return;
	}
	STATISTICS_INC(m_Statistics.m_DeallocateRequest);
	
	CHUNK* chunk = (CHUNK*)((char*)ptr - CHUNK_HEADER_SIZE);
	PAFEE_ASSERT(chunk);
	if (chunk->m_SanityCode != CHUNK_SANITY_CODE)
	{
#ifdef PAFEE_EXCEPTION
		throw std::runtime_error("demaged chunk");
#endif
		return;
	}

	PAGE* page = (PAGE*)((char*)chunk - chunk->m_PageOffset);
	PAFEE_ASSERT(page);
	if (page->m_SanityCode != PAGE_SANITY_CODE)
	{
#ifdef PAFEE_EXCEPTION
		throw std::runtime_error("demaged page");
#endif
		return;
	}

	STATISTICS_SUB(m_Statistics.m_UsingMemorySizeAll, m_ChunkSize);

	DeallocateChunk(page, ptr, m_ChunkSize, m_Chunks);
}

//--------------------------------------------------------------------
//!  Check given address is in own address range
//!  @return: true: own address, false: not own address
//--------------------------------------------------------------------
bool FixedAllocator::IsOwnAddress(void* ptr)
{
	unsigned long m_PageSize = (m_ChunkSize*m_Chunks);	//@@@
	PAGE* pPage = m_PageList;
	while(pPage)
	{
		if((pPage <= ptr) && (ptr < (pPage+m_PageSize)))
		{
			return true;
		}
		pPage = pPage->m_pNext;
	}
	return false;
}


#ifdef PAFEE_STATISTICS_ENABLE
//--------------------------------------------------------------------
//!  get statistice information
//!  @return: void
//--------------------------------------------------------------------
void FixedAllocator::GetStatistics(
		FixedAllocator::STATISTICS* pStat ///< [in/out] statistice information pointer
		)
{
	pafee_memcpy(pStat, &m_Statistics, sizeof(m_Statistics));
}
#endif

//--------------------------------------------------------------------
//!  get instance
//!  @return: pointer for instance
//--------------------------------------------------------------------
FixedAllocator* FixedAllocator::GetInstance(
		void* ptr ///< [in] allocated memory pointer
		)
{
	if (ptr == NULL)
	{
		return NULL;
	}

	CHUNK* chunk = (CHUNK*)((char*)ptr - CHUNK_HEADER_SIZE);
	if (chunk->m_SanityCode != CHUNK_SANITY_CODE)
	{
		return NULL;
	}

	PAGE* page = (PAGE*)((char*)chunk - chunk->m_PageOffset);
	if (page->m_SanityCode != PAGE_SANITY_CODE)
	{
		return NULL;
	}

	return (FixedAllocator*)page->m_pThis;
}

