//********************************************************************
//		project 	PAFEE
//
//! 	 @file: 	PafeeFixedAllocator.h
//! 	 @brief:	fixed memory allocation
//
//********************************************************************

//
// P A F E E   - Pragmatic Application Framework for Embedded Environment- 
//
//
// Copyright (c) 2009 KN Information Sustems Corporation,
//                    Hirocon Corporation,
//                    Technocrats Corporation
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//

#ifndef _FIXED_ALLOCATOR_H_INCLUDED_
#define _FIXED_ALLOCATOR_H_INCLUDED_

#include <stddef.h>
#include "PafeeCommon.h"
#include "PafeeAllocator.h"


/* declare public C functions */
#ifdef __cplusplus
namespace Pafee {

	//
	// class definition
	//
	//--------------------------------------------------------------------
	//  FixedAllocator class
	//--------------------------------------------------------------------
	class FixedAllocator : public Allocator
	{
	public:
#ifdef PAFEE_STATISTICS_ENABLE
		//! statistics information
		typedef struct _STATISTICS
		{ 
			unsigned int m_AllocateRequest;
			unsigned int m_DeallocateRequest;
			unsigned int m_GetHeap;
			unsigned int m_ReleaseHeap;
			unsigned long m_UsingMemorySizeAll;
			unsigned long m_UsingPageSizeAll;
			unsigned long m_RequestMemorySizeMin;
			unsigned long m_RequestMemorySizeMax;
			unsigned int m_RequestSizeUnder20PerChunk;
			unsigned int m_RequestSizeUnder40PerChunk;
			unsigned int m_RequestSizeUnder60PerChunk;
			unsigned int m_RequestSizeUnder80PerChunk;
			unsigned int m_RequestSizeUnder100PerChunk;
		} STATISTICS;
#endif

	protected:

		//! chunk information
		typedef struct _CHUNK { 
			unsigned long m_SanityCode;	///< sanity code
			unsigned long m_PageOffset;	///< page offset
			unsigned long m_Data[1];	///< chunk data
		} CHUNK;

		//! page information
		typedef struct _PAGE {
			unsigned short m_SanityCode; 			///< sanity code
			unsigned short m_FirstAvailableChunk; 	///< first available chunk
			void* m_pThis;  						///< pointer for instance
			struct _PAGE* m_pNext;  				///< pointer for next page
			struct _PAGE* m_pPrevious; 				///< pointer for previous page
			char m_Data[1]; 						///< page pointer
		} PAGE;


		PAGE* m_AllocPage;				///< current page
		PAGE* m_PageList; 				///< pointer for page list
		unsigned long m_ChunkSize; 		///< chunk size
		unsigned long m_Chunks; 		///< amount of chunk
		Allocator* m_pBackendAllocator;	///< backend allocator

#ifdef PAFEE_STATISTICS_ENABLE
		STATISTICS m_Statistics;
#endif

		static const unsigned long  CHUNK_HEADER_SIZE;	///< chunk header size
		static const unsigned long  PAGE_HEADER_SIZE;	///< page header size
		static const unsigned short PAGE_SANITY_CODE;	///< page sanity check code
		static const unsigned long  CHUNK_SANITY_CODE;	///< chunk sanity check code
		
	public:
		//constractors and destoractor
		FixedAllocator();
		virtual ~FixedAllocator();

		//other methods
		bool Initialize(unsigned long chunkSize, unsigned long chunkAmount, unsigned long initialPages, Allocator* pBackendAllocator);
		virtual void* Allocate(unsigned long size);
		virtual void Deallocate(void* ptr);
		virtual bool IsOwnAddress(void* ptr);
		inline unsigned long GetAllocateSize(void) { return m_ChunkSize; }
		static FixedAllocator* GetInstance(void* ptr);

#ifdef PAFEE_STATISTICS_ENABLE
		void GetStatistics(FixedAllocator::STATISTICS* pStat);
#endif

	protected:
		PAGE* IntializePage(unsigned long chunkSize, unsigned long chunks);
		void* AllocateChunk(PAGE* page, unsigned long chunkSize, unsigned long chunks);
		bool DeallocateChunk(PAGE* page, void* ptr, unsigned long chunkSize, unsigned long chunks);
#ifdef PAFEE_STATISTICS_ENABLE
		void CalcStatistics(STATISTICS* pStat, unsigned long requestSize, unsigned long defaultSize);
#endif
	private:
		FixedAllocator(const FixedAllocator& rhs);
		FixedAllocator& operator=(const FixedAllocator& rhs);
	};

};
#endif

#endif // _FIXED_ALLOCATOR_H_INCLUDED_
