//********************************************************************
//
//! 	 @file: 	PafeeFlexibleAllocator.h
//! 	 @author:	T.Miyawaki
//! 	 @brief:	Flexible allocator
//
//********************************************************************

//
// P A F E E   - Pragmatic Application Framework for Embedded Environment- 
//
//
// Copyright (c) 2009 KN Information Sustems Corporation,
//                    Hirocon Corporation,
//                    Technocrats Corporation
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//

#ifndef _FLEXIBLE_ALLOCATOR_H_INCLUDED_
#define _FLEXIBLE_ALLOCATOR_H_INCLUDED_

#include "PafeeCommon.h"
#include "PafeeAllocator.h"

namespace Pafee
{

class FlexibleAllocator : public Allocator
{

public:
	//! Constractor
	FlexibleAllocator();
	
	//! Destoractor
	virtual ~FlexibleAllocator();
	
	//! Initialize (Allocate heap by given allocator)
	bool Initialize(
				unsigned long heapSize,
				unsigned long maxAllocateSize,
				unsigned long minBlockSize,
				Allocator* pAllocator=NULL
				);
	
	//! Initialize (Use given heap)
	void Initialize(
				unsigned long heapAddr,
				unsigned long heapSize,
				unsigned long maxAllocateSize,
				unsigned long minBlockSize
				);
	
	//! Allocate memory
	virtual void* Allocate(unsigned long size);

	//! Deallocate memory
	virtual void Deallocate(void* ptr);

	//! Check given address is in own address range
	virtual bool IsOwnAddress(void* ptr);

#ifdef PAFEE_STATISTICS_ENABLE
	//! Statistics
	typedef struct
	{
		unsigned long allocateRequest;			///< Allocation request count
		unsigned long deallocateRequest;		///< Dealocation request count
		unsigned long allocateFailure;			///< Allocation failure count
		unsigned long sanityErrors;				///< Sanity check error count
		
		unsigned long maxUsedBytes;				///< max used memory size
		unsigned long minFreeBytes;				///< min free memory size
		unsigned long maxAllocateSize;			///< max allocation request size
		unsigned long minAllocateSize;			///< min allocation request size
		unsigned long maxAllocateSeeks;			///< max block seek count on allocation
		
		//! current statistics
		struct
		{
			unsigned long usingBytes;				///< using memory size
			unsigned long freeBytes;				///< free memory size
			unsigned long freeListElements;			///< free-list element count
			unsigned long usingListElements;		///< using-list element count
			unsigned long minBlockSizeInFreeList;	///< min block size in free-list
			unsigned long maxBlockSizeInFreeList;	///< max block size in free-list
			unsigned long minBlockSizeInUsingList;	///< max block size in using-list
			unsigned long maxBlockSizeInUsingList;	///< min block size in using-list
			unsigned long sanityErrorInFreeList;	///< sanity error address in free-list
			unsigned long sanityErrorInUsingList;	///< sanity error address in using-list
			
			unsigned long availableMaxSizeBlocks;	///< available max size allocation
			unsigned long available1_2SizeBlocks;	///< available 1/2 size allocation
			unsigned long available1_4SizeBlocks;	///< available 1/4 size allocation
			unsigned long available1_8SizeBlocks;	///< available 1/8 size allocation
		}current;
	} STATISTICS;

	//! Get statistics
	void GetStatistics(FlexibleAllocator::STATISTICS* pStat);

	//! Get statistics
	virtual void GetStatistics(StreamOutput& stream);
#endif
#ifdef PAFEE_UNITTEST
	//! Unit tes
	static void UnitTest(void);
	void ShowLists(void);
#endif


protected:
	static const unsigned long SANITY_CODE;	///< Sanity code
	static const unsigned long USING_MARK;	///< using mark

	
	//! Block header information
	typedef struct _BLK_HEADER
	{
		unsigned long	sanity_u;		///< Sanity-code(bit31-1), Using-flag(bit0)
		unsigned long	blockSize;		///< Block size
		_BLK_HEADER*	pFrontBlock;	///< front block in memory layout
		_BLK_HEADER*	pBackBlock;		///< back block in memory layout
		_BLK_HEADER*	pPrev;			///< previous block in list
		_BLK_HEADER*	pNext;			///< next block in list
	} BLK_HEADER;
	
	
	unsigned long m_heapAddr;			///< start address of  heap
	unsigned long m_heapSize;			///< heap size
	unsigned long m_maxAllocateSize;	///< maximum allocate size
	unsigned long m_minBlockSize;		///< minimum block size
	unsigned long m_endAddr;			///< end address
	
	BLK_HEADER*	m_pFreeListHead;		///< head element of free-list
	BLK_HEADER*	m_pUsingListHead;		///< head element of using-list
	Allocator*	m_pAllocator;			///< Allocator to allocate heap
	bool		m_isOwnHeapAllocation;	///< Own heap allocation flag
	
#ifdef PAFEE_STATISTICS_ENABLE
	STATISTICS	m_stat;					///< m_statistics
#endif

private:
	//! Copy constractor
	FlexibleAllocator(const FlexibleAllocator& src);
	
	//! Operator equal
	FlexibleAllocator& operator=(const FlexibleAllocator& src);
};

};

#endif // _FLEXIBLE_ALLOCATOR_H_INCLUDED_
