//********************************************************************
//		project 	PAFEE
//
//! 	 @file: 	PafeeSemaphore.cpp
//! 	 @author:	Technocrats
//! 	 @brief:	PafeeSemaphore class
//
//********************************************************************
//
// P A F E E   - Pragmatic Application Framework for Embedded Environment- 
//
//
// Copyright (c) 2009 KN Information Sustems Corporation,
//                    Hirocon Corporation,
//                    Technocrats Corporation
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//

// if no OS definition, PafeeSemaphore class is unavailable
#if defined(PAFEE_OS_WINDOWS)||defined(PAFEE_OS_LINUX)||defined(PAFEE_OS_ITRON)

#include <limits.h>
#include <stdarg.h>
#include "PafeeSemaphore.h"

#ifdef __cplusplus
namespace Pafee
{
#endif

#if defined(PAFEE_OS_ITRON)&&defined(PAFEE_DEBUG)
/**
 * @brief  Log output function(for ITRON simulator on Windows)
 * @param  *format, ... : Log contents
 * @return (none)
 */
void SemaphoreSyslog(
	const char *format, ...
)
{
	const char *string;
	int data;

	va_list list;
	va_start(list, format);
	string = format;
	data = va_arg(list, int);
	syslog(LOG_NOTICE, format, data);
}
#endif // PAFEE_OS_ITRON && PAFEE_DEBUG

/**
 * @brief  PafeeSemaphore class constructor
 * @param  (none)
 * @return (none)
 */
PafeeSemaphore::PafeeSemaphore(
)
{
	SEM_DEBUG((">> PafeeSemaphore::PafeeSemaphore\n"));
#ifdef PAFEE_OS_WINDOWS
	m_handle = NULL;
#endif
#ifdef PAFEE_OS_LINUX
//	m_handle = 0;
	m_isCreated = false;
#endif
#ifdef PAFEE_OS_ITRON
	m_handle = 0;
#endif
	m_error = ERROR_NONE;
	SEM_DEBUG(("<< PafeeSemaphore::PafeeSemaphore\n"));
}

/**
 * @brief  PafeeSemaphore class destructor
 * @param  (none)
 * @return (none)
 */
PafeeSemaphore::~PafeeSemaphore(
)
{
	SEM_DEBUG((">> PafeeSemaphore::~PafeeSemaphore\n"));
	Delete();
	SEM_DEBUG(("<< PafeeSemaphore::~PafeeSemaphore\n"));
}

#if defined(PAFEE_OS_WINDOWS)
/**
 * @brief  Make new semaphore(for Windows)
 * @param  (none)
 * @return Create success or not
 * @retval true:create succeeded
 * @retval false:create failed
 */
bool PafeeSemaphore::Create(
	int maxCount ///< [in] Max count of semaphore
)
{
	SEM_DEBUG((">> PafeeSemaphore::Create\n"));
	if(m_handle != NULL)
	{
		m_error = ERROR_HAS_HANDLE;
		SEM_DEBUG(("<<FAIL1 PafeeSemaphore::Create\n"));
		return false;
	}

	m_handle = CreateSemaphore(NULL, maxCount, maxCount, NULL); // no name semaphore
	if(m_handle == NULL)
	{
		m_error = ERROR_CANNOT_CREATE;
		SEM_DEBUG(("<<FAIL2 PafeeSemaphore::Create\n"));
		return false;
	}

	SEM_DEBUG(("<< PafeeSemaphore::Create\n"));
	return true;
}
#endif
#ifdef PAFEE_OS_LINUX
/**
 * @brief  Make new semaphore(for Linux)
 * @param  (none)
 * @return Create success or not
 * @retval true:create succeeded
 * @retval false:create failed
 */
bool PafeeSemaphore::Create(
	int maxCount ///< [in] Max count of semaphore
)
{
	SEM_DEBUG((">> PafeeSemaphore::Create\n"));
	if(m_isCreated == true)
	{
		m_error = ERROR_HAS_HANDLE;
		SEM_DEBUG(("<<FAIL1 PafeeSemaphore::Create\n"));
		return false;
	}
	// initialize semaphore (shared between threads in the process)
	sem_init(&m_handle, 0, maxCount);

	m_isCreated = true;
	SEM_DEBUG(("<< PafeeSemaphore::Create\n"));
	return true;
}
#endif

#ifdef PAFEE_OS_ITRON
/**
 * @brief  Attach static semaphore id(for ITRON)
 * @return Attach success or not
 * @retval true:create succeeded
 * @retval false:create failed
 */
bool PafeeSemaphore::Attach(
	ID id ///< [in] Semaphore ID
)
{
	SEM_DEBUG((">> PafeeSemaphore::Attach\n"));
	if(m_handle != 0)
	{
		m_error = ERROR_HAS_HANDLE;
		SEM_DEBUG(("<<FAIL1 PafeeSemaphore::Attach\n"));
		return false;
	}

	m_handle = id;
	SEM_DEBUG(("<< PafeeSemaphore::Attach\n"));
	return true;
}
#endif

#ifdef PAFEE_OS_WINDOWS
/**
 * @brief  Delete semaphore(for Windows)
 * @param  (none)
 * @return Delete success or not
 * @retval true:Delete succeeded
 * @retval false:Delete failed
 */
bool PafeeSemaphore::Delete(
)
{
	int returnValue;

	SEM_DEBUG((">> PafeeSemaphore::Delete\n"));
	if(m_handle == NULL)
	{
		m_error = ERROR_NOHANDLE;
		SEM_DEBUG(("<<FAIL1 PafeeSemaphore::Delete\n"));
		return false;
	}
	else
	{
		returnValue = CloseHandle(m_handle);

		if(returnValue == 0)
		{
			m_error = ERROR_CANNOT_DELETE;
			SEM_DEBUG(("<<FAIL2 PafeeSemaphore::Delete\n"));
			return false;
		}
		else
		{
			m_handle = NULL;
			SEM_DEBUG(("<< PafeeSemaphore::Delete\n"));
			return true;
		}
	}
}
#endif
#ifdef PAFEE_OS_LINUX
/**
 * @brief  Delete semaphore(for Linux)
 * @param  (none)
 * @return Delete success or not
 * @retval true:Delete succeeded
 * @retval false:Delete failed
 */
bool PafeeSemaphore::Delete(
)
{
	int returnValue;

	SEM_DEBUG((">> PafeeSemaphore::Delete\n"));
	if(m_isCreated == false)
	{
		m_error = ERROR_NOHANDLE;
		SEM_DEBUG(("<<FAIL1 PafeeSemaphore::Delete\n"));
		return false;
	}

	returnValue = sem_destroy(&m_handle);

	if(returnValue != 0)
	{
		m_error = ERROR_CANNOT_DELETE;
		SEM_DEBUG(("<<FAIL2 PafeeSemaphore::Delete\n"));
		return false;
	}
	else
	{
		m_isCreated = false;
		SEM_DEBUG(("<< PafeeSemaphore::Delete\n"));
		return true;
	}
}
#endif
#ifdef PAFEE_OS_ITRON
/**
 * @brief  Free static semaphore id(for ITRON)
 * @param  (none)
 * @return Delete success or not
 * @retval true:Delete succeeded
 * @retval false:Delete failed
 */
bool PafeeSemaphore::Delete(
)
{
//	ER returnValue;

	SEM_DEBUG((">> PafeeSemaphore::Delete\n"));

#if 0	// no way to delete semaphore with Standard Profile
	returnValue = del_sem(m_handle);

	if(returnValue != E_OK)
	{
		m_error = ERROR_CANNOT_DELETE;
		SEM_DEBUG(("<<FAIL1 PafeeSemaphore::Delete\n"));
		return false;
	}
	else
	{
		m_handle = 0;
		SEM_DEBUG(("<< PafeeSemaphore::Delete\n"));
		return true;
	}
#else
	m_handle = 0;
	SEM_DEBUG(("<< PafeeSemaphore::Delete\n"));
	return true;
#endif
}
#endif

#ifdef PAFEE_OS_WINDOWS
/**
 * @brief  Take semaphore. If can not take, wait until take(for Windows)
 * @param  (none)
 * @return Take success or not
 * @retval true:Take succeeded
 * @retval false:Take failed
 */
bool PafeeSemaphore::Take(
)
{
	unsigned long errorCode;

	SEM_DEBUG((">> PafeeSemaphore::Take\n"));
	if(m_handle == NULL)
	{
		m_error = ERROR_NOHANDLE;
		SEM_DEBUG(("<<FAIL1 PafeeSemaphore::Take\n"));
		return false;
	}
	else
	{
		errorCode = WaitForSingleObject(m_handle, INFINITE);
		switch(errorCode)
		{
		case WAIT_OBJECT_0:
		case WAIT_ABANDONED:
			SEM_DEBUG(("<< PafeeSemaphore::Take\n"));
			return true;
		case WAIT_TIMEOUT:
			m_error = ERROR_TIME_OUT;
			SEM_DEBUG(("<<FAIL2 PafeeSemaphore::Take\n"));
			return false;
		case WAIT_FAILED:
		default:
			m_error = ERROR_CANNOT_GET;
			SEM_DEBUG(("<<FAIL2 PafeeSemaphore::Take\n"));
			return false;
		}
	}
}
#endif
#ifdef PAFEE_OS_LINUX
/**
 * @brief  Take semaphore. If can not take, wait until take(for Linux)
 * @param  (none)
 * @return Take success or not
 * @retval true:Take succeeded
 * @retval false:Take failed
 */
bool PafeeSemaphore::Take(
)
{
	SEM_DEBUG((">> PafeeSemaphore::Take\n"));
	if(m_isCreated == false)
	{
		m_error = ERROR_NOHANDLE;
		SEM_DEBUG(("<<FAIL1 PafeeSemaphore::Take\n"));
		return false;
	}
	else
	{
		sem_wait(&m_handle); // always return 0
		SEM_DEBUG(("<< PafeeSemaphore::Take\n"));
		return true;
	}
}
#endif
#ifdef PAFEE_OS_ITRON
/**
 * @brief  Take semaphore. If can not take, wait until take(for ITRON)
 * @param  (none)
 * @return Take success or not
 * @retval true:Take succeeded
 * @retval false:Take failed
 */
bool PafeeSemaphore::Take(
)
{
	ER returnValue;

	SEM_DEBUG((">> PafeeSemaphore::Take\n"));
	if(m_handle == 0)
	{
		m_error = ERROR_NOHANDLE;
		SEM_DEBUG(("<<FAIL1 PafeeSemaphore::Take\n"));
		return false;
	}

	returnValue = wai_sem(m_handle);
	if(returnValue != E_OK)
	{
		m_error = ERROR_CANNOT_GET;
		SEM_DEBUG(("<<FAIL2 PafeeSemaphore::Take\n"));
		return false;
	}
	else
	{
		SEM_DEBUG(("<< PafeeSemaphore::Take\n"));
		return true;
	}
}
#endif

#ifdef PAFEE_OS_WINDOWS
/**
 * @brief  Asynchronized take. If can not take, return false(for Windows)
 * @param  (none)
 * @return Take success or not
 * @retval true:Take succeeded
 * @retval false:Take failed
 */
bool PafeeSemaphore::AsyncTake(
)
{
	unsigned long errorCode;

	SEM_DEBUG((">> PafeeSemaphore::AsyncTake\n"));
	if(m_handle == NULL)
	{
		m_error = ERROR_NOHANDLE;
		SEM_DEBUG(("<<FAIL1 PafeeSemaphore::Take\n"));
		return false;
	}
	else
	{
		errorCode = WaitForSingleObject(m_handle, 0);

		switch(errorCode)
		{
		case WAIT_OBJECT_0:
		case WAIT_ABANDONED:
			SEM_DEBUG(("<< PafeeSemaphore::AsyncTake\n"));
			return true;
		case WAIT_TIMEOUT:
			m_error = ERROR_TIME_OUT;
			SEM_DEBUG(("<<FAIL2 PafeeSemaphore::AsyncTake\n"));
			return false;
		case WAIT_FAILED:
		default:
			m_error = ERROR_CANNOT_GET;
			SEM_DEBUG(("<<FAIL3 PafeeSemaphore::AsyncTake\n"));
			return false;
		}
	}
}
#endif
#ifdef PAFEE_OS_LINUX
/**
 * @brief  Asynchronized take. If can not take, return false(for Linux)
 * @param  (none)
 * @return Take success or not
 * @retval true:Take succeeded
 * @retval false:Take failed
 */
bool PafeeSemaphore::AsyncTake(
)
{
	int returnValue;

	SEM_DEBUG((">> PafeeSemaphore::AsyncTake\n"));
	if(m_isCreated == false)
	{
		m_error = ERROR_NOHANDLE;
		SEM_DEBUG(("<<FAIL1 PafeeSemaphore::AsyncTake\n"));
		return false;
	}
	else
	{
		returnValue = sem_trywait(&m_handle);

		if(returnValue != 0)
		{
			m_error = ERROR_CANNOT_GET;
			SEM_DEBUG(("<<FAIL2 PafeeSemaphore::AsyncTake\n"));
			return false;
		}
		else
		{
			SEM_DEBUG(("<< PafeeSemaphore::AsyncTake\n"));
			return true;
		}
	}
}
#endif
#ifdef PAFEE_OS_ITRON
/**
 * @brief  Asynchronized take. If can not take, return false(for ITRON)
 * @param  (none)
 * @return Take success or not
 * @retval true:Take succeeded
 * @retval false:Take failed
 */
bool PafeeSemaphore::AsyncTake(
)
{
	ER returnValue;

	SEM_DEBUG((">> PafeeSemaphore::AsyncTake\n"));
	if(m_handle == 0)
	{
		m_error = ERROR_NOHANDLE;
		SEM_DEBUG(("<<FAIL1 PafeeSemaphore::AsyncTake\n"));
		return false;
	}

	returnValue = pol_sem(m_handle);
	if(returnValue != E_OK)
	{
		switch(returnValue)
		{
		case E_TMOUT:
			m_error = ERROR_TIME_OUT;
			SEM_DEBUG(("<<FAIL2 PafeeSemaphore::AsyncTake\n"));
			return false;
		default:
			m_error = ERROR_CANNOT_GET;
			SEM_DEBUG(("<<FAIL3 PafeeSemaphore::AsyncTake\n"));
			return false;
		}
	}
	else
	{
		SEM_DEBUG(("<< PafeeSemaphore::AsyncTake\n"));
		return true;
	}
}
#endif

#if defined(PAFEE_OS_WINDOWS)
/**
 * @brief  Timed take. If can not take, wait until time is up(for Windows)
 * @return Take success or not
 * @return Take success or not
 * @retval true:Take succeeded
 * @retval false:Take failed
 */
bool PafeeSemaphore::TimedTake(
	int time ///< [in] Timeout time
)
{
	unsigned long errorCode;

	SEM_DEBUG((">> PafeeSemaphore::TimedTake\n"));
	if(m_handle == NULL)
	{
		m_error = ERROR_NOHANDLE;
		SEM_DEBUG(("<<FAIL1 PafeeSemaphore::TimedTake\n"));
		return false;
	}
	else
	{
		errorCode = WaitForSingleObject(m_handle, time); // unit of time : ms

		switch(errorCode)
		{
		case WAIT_OBJECT_0:
		case WAIT_ABANDONED:
			SEM_DEBUG(("<< PafeeSemaphore::TimedTake\n"));
			return true;
		case WAIT_TIMEOUT:
			m_error = ERROR_TIME_OUT;
			SEM_DEBUG(("<<FAIL2 PafeeSemaphore::TimedTake\n"));
			return false;
		case WAIT_FAILED:
		default:
			m_error = ERROR_CANNOT_GET;
			SEM_DEBUG(("<<FAIL3 PafeeSemaphore::TimedTake\n"));
			return false;
		}
	}
}
#endif
#if defined(PAFEE_OS_LINUX)
/**
 * @brief  Timed take. If can not take, wait until time is up(for Linux)
 * @return Take success or not
 * @return Take success or not
 * @retval true:Take succeeded
 * @retval false:Take failed
 */
bool PafeeSemaphore::TimedTake(
	int time ///< [in] Timeout time
)
{
#if 0
	int returnValue;
	struct timeval timeValue;
	unsigned long startTime, currentTime;

	SEM_DEBUG((">> PafeeSemaphore::TimedTake\n"));
	if(m_isCreated == false)
	{
		m_error = ERROR_NOHANDLE;
		SEM_DEBUG(("<<FAIL1 PafeeSemaphore::TimedTake\n"));
		return false;
	}
	else
	{
		gettimeofday(&timeValue, 0);
		// unit of time : ms
		startTime = timeValue.tv_sec*1000 + timeValue.tv_usec/1000;

		while(1)
		{
			returnValue = sem_trywait(&m_handle);
			if(returnValue != 0)
			{
				gettimeofday(&timeValue, 0);
				currentTime = timeValue.tv_sec*1000 + timeValue.tv_usec/1000;

				if((currentTime - startTime) > time)
				{
					m_error = ERROR_TIME_OUT;
					SEM_DEBUG(("<<FAIL2 PafeeSemaphore::TimedTake\n"));
					return false;
				}
			}
			else
			{
				SEM_DEBUG(("<< PafeeSemaphore::TimedTake\n"));
				return true;
			}
		} // end of while roop
	}
#else
	int returnValue;
	struct timespec endTime;

	SEM_DEBUG((">> PafeeSemaphore::TimedTake\n"));
	if(m_isCreated == false)
	{
		m_error = ERROR_NOHANDLE;
		SEM_DEBUG(("<<FAIL1 PafeeSemaphore::TimedTake\n"));
		return false;
	}
	else
	{
		returnValue = clock_gettime(CLOCK_REALTIME, &endTime);
		if(returnValue != 0)
		{
			m_error = ERROR_CANNOT_GET;
			SEM_DEBUG(("<<FAIL2 PafeeSemaphore::TimedTake\n"));
			return false;
		}

		// unit of time : ms
		endTime.tv_sec += time/1000;
		endTime.tv_nsec += (time%1000)*1000000;
		if(endTime.tv_nsec > 1000000000)  // nsec value is over 1sec
		{
			endTime.tv_nsec -= 1000000000;
			endTime.tv_sec += 1;
		}

		returnValue = sem_timedwait(&m_handle, &endTime);
		if(returnValue != 0)
		{
			if(errno == ETIMEDOUT)
			{
				m_error = ERROR_TIME_OUT;
				SEM_DEBUG(("<<FAIL3 PafeeSemaphore::TimedTake\n"));
				return false;
			}
			else
			{
				m_error = ERROR_CANNOT_GET;
				SEM_DEBUG(("<<FAIL4 PafeeSemaphore::TimedTake\n"));
				return false;
			}
		}
		else
		{
			SEM_DEBUG(("<< PafeeSemaphore::TimedTake\n"));
			return true;
		}
	}
#endif
}
#endif
#if defined(PAFEE_OS_ITRON)
/**
 * @brief  Timed take. If can not take, wait until time is up(for ITRON)
 * @return Take success or not
 * @return Take success or not
 * @retval true:Take succeeded
 * @retval false:Take failed
 */
bool PafeeSemaphore::TimedTake(
	int time ///< [in] Timeout time
)
{
	ER returnValue;
	SEM_DEBUG((">> PafeeSemaphore::TimedTake\n"));

	if(m_handle == 0)
	{
		m_error = ERROR_NOHANDLE;
		SEM_DEBUG(("<<FAIL1 PafeeSemaphore::TimedTake\n"));
		return false;
	}

	// unit of time : depending on the system
	returnValue = twai_sem(m_handle, time);
	if(returnValue != E_OK)
	{
		switch(returnValue)
		{
		case E_TMOUT:
			m_error = ERROR_TIME_OUT;
			SEM_DEBUG(("<<FAIL2 PafeeSemaphore::TimedTake\n"));
			return false;
		default:
			m_error = ERROR_CANNOT_GET;
			SEM_DEBUG(("<<FAIL3 PafeeSemaphore::TimedTake\n"));
			return false;
		}
	}
	else
	{
		SEM_DEBUG(("<< PafeeSemaphore::TimedTake\n"));
		return true;
	}
}
#endif

#if defined(PAFEE_OS_WINDOWS)
/**
 * @brief  Release semaphore. If can not release, return false(for Windows)
 * @param  (none)
 * @return Release success or not
 * @retval true:Release succeeded
 * @retval false:Release failed
 */
bool PafeeSemaphore::Release(
)
{
	int returnValue;

	SEM_DEBUG((">> PafeeSemaphore::Release\n"));
	if(m_handle == NULL)
	{
		m_error = ERROR_NOHANDLE;
		SEM_DEBUG(("<<FAIL1 PafeeSemaphore::Release\n"));
		return false;
	}
	else
	{
		returnValue = ReleaseSemaphore(m_handle, 1, NULL);

		if(returnValue == 0)
		{
			m_error = ERROR_CANNOT_RELEASE;
			SEM_DEBUG(("<<FAIL2 PafeeSemaphore::Release\n"));
			return false;
		}
		else
		{
			SEM_DEBUG(("<< PafeeSemaphore::Release\n"));
			return true;
		}
	}
}
#endif
#if defined(PAFEE_OS_LINUX)
/**
 * @brief  Release semaphore. If can not release, return false(for Linux)
 * @param  (none)
 * @return Release success or not
 * @retval true:Release succeeded
 * @retval false:Release failed
 */
bool PafeeSemaphore::Release(
)
{
	int returnValue;
	SEM_DEBUG((">> PafeeSemaphore::Release\n"));

	if(m_isCreated == false)
	{
		m_error = ERROR_NOHANDLE;
		SEM_DEBUG(("<<FAIL1 PafeeSemaphore::Release\n"));
		return false;
	}
	else
	{
		returnValue = sem_post(&m_handle);

		if(returnValue != 0)
		{
			m_error = ERROR_CANNOT_RELEASE;
			SEM_DEBUG(("<<FAIL2 PafeeSemaphore::Release\n"));
			return false;
		}
		else
		{
			SEM_DEBUG(("<< PafeeSemaphore::Release\n"));
			return true;
		}
	}
}
#endif
#if defined(PAFEE_OS_ITRON)
/**
 * @brief  Release semaphore. If can not release, return false(for ITRON)
 * @param  (none)
 * @return Release success or not
 * @retval true:Release succeeded
 * @retval false:Release failed
 */
bool PafeeSemaphore::Release(
)
{
	ER returnValue;
	SEM_DEBUG((">> PafeeSemaphore::Release\n"));

	if(m_handle == 0)
	{
		m_error = ERROR_NOHANDLE;
		SEM_DEBUG(("<<FAIL1 PafeeSemaphore::Release\n"));
		return false;
	}

	returnValue = sig_sem(m_handle);
	if(returnValue != E_OK)
	{
		m_error = ERROR_CANNOT_RELEASE;
		SEM_DEBUG(("<<FAIL2 PafeeSemaphore::Release\n"));
		return false;
	}
	else
	{
		SEM_DEBUG(("<< PafeeSemaphore::Release\n"));
		return true;
	}
}
#endif

/**
 * @brief  Get error code. 
 * @param  (none)
 * @return Semaphore error code
 * @retval SemaphoreError
 */
PafeeSemaphore::SEM_ERROR PafeeSemaphore::GetLastError(
)
{
	SEM_DEBUG((">> PafeeSemaphore::GetLastError\n"));
	SEM_DEBUG(("<< PafeeSemaphore::GetLastError\n"));
	return m_error;
}

#ifdef __cplusplus
}; // namespace pafee
#endif

#endif // PAFEE_OS_WINDOWS || PAFEE_OS_LINUX || PAFEE_OS_ITRON
