//********************************************************************
//		project 	PAFEE
//
//! 	 @file: 	Signal.h
//! 	 @author:	Technocrats
//! 	 @brief:	Signal class header
//
//********************************************************************
//
// P A F E E   - Pragmatic Application Framework for Embedded Environment- 
//
//
// Copyright (c) 2009 KN Information Sustems Corporation,
//                    Hirocon Corporation,
//                    Technocrats Corporation
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND noneINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//

#ifndef _PAFEE_SIGNAL_H_INCLUDED_
#define _PAFEE_SIGNAL_H_INCLUDED_

#ifdef PAFEE_OS_WINDOWS
#include <windows.h>
#endif

#ifdef PAFEE_OS_LINUX
#include <pthread.h>
#endif

#ifdef PAFEE_OS_ITRON
#include <kernel.h>
#endif

#ifdef PAFEE_DEBUG
#include <stdio.h>
#endif

namespace Pafee{

#ifdef PAFEE_DEBUG
#ifdef PAFEE_OS_ITRON
#define SIGNAL_DEBUG(STRING) SignalSyslog STRING
#else
#define	SIGNAL_DEBUG(STRING) printf STRING
#endif
#else
#define SIGNAL_DEBUG(STRING)
#endif

#ifdef PAFEE_OS_ITRON
int SignalSyslog(const char *string, ...);
#endif

#if defined(PAFEE_OS_WINDOWS) || defined(PAFEE_OS_LINUX) || defined(PAFEE_OS_ITRON)
class PafeeSignal{
public:
	/**
	 * @brief  Signal error emun
	 */
	typedef enum tagSIGNAL_ERROR{
		ERROR_NONE, /**< Error none */
		ERROR_NOSIGNAL, /**< handle, id is not exist */
		ERROR_HAS_SIGNAL, /**< Allready exist m_handle */
		ERROR_CANNOT_CREATE, /**< Signal can not create */
		ERROR_CANNOT_DELETE, /**< Signal can not delete */
		ERROR_CANNOT_SIGNAL, /**< Signal can not get */
		ERROR_TIME_OUT, /**< Signal can not get for time out */
		ERROR_CANNOT_SET_EVENT, /**< Can not send signal event */
		ERROR_CANNOT_RESET_EVENT, /**< Can not release signal event */
		ERROR_CANNOT_CLEAR, /**< Can not clear event */
		ERROR_OHTER, /**< Unknown error */
		ERROR_MAX /**< Error none */
	} SIGNAL_ERROR;

private:
#ifdef PAFEE_OS_WINDOWS
	HANDLE m_handle; /**< Windows signal handle */
#endif // PAFEE_OS_WINDOWS

#ifdef PAFEE_OS_LINUX
	pthread_cond_t m_condition; /**< Linux signal condition */
	pthread_mutex_t m_mutex; /**< Linux mutex handle */
	bool m_signal; /**< Send Signal flag */
	bool m_created; /**< signal handle created flag */
#endif // PAFEE_OS_LINUX

#ifdef PAFEE_OS_ITRON
	ID m_id; /**< Itron mutex handle */
#endif // PAFEE_OS_ITRON

	SIGNAL_ERROR m_error;

public:
	/**
	 * @brief  PafeeSignal class constructor
	 * @param  none
	 * @return none
	 */
	PafeeSignal();

	/**
	 * @brief  PafeeSignal class destructor
	 * @param  none
	 * @return none
	 */
	~PafeeSignal();

#if defined(PAFEE_OS_WINDOWS) || defined(PAFEE_OS_LINUX)
	/**
	 * @brief  Make new signal
	 * @param  none
	 * @return Create success or not
	 * @retval true:create success
	 * @retval false:create failed
	 */
	bool Create();
#endif //  defined(PAFEE_OS_WINDOWS) || defined(PAFEE_OS_LINUX)

#ifdef PAFEE_OS_ITRON
	/**
	 * @brief  Attach static signal id
	 * @return Attach success or not
	 * @retval true:create success
	 * @retval false:create failed
	 */
	bool Attach(
		ID id ///< [in] Signal ID
	);
#endif // PAFEE_OS_ITRON

	/**
	 * @brief  Delete m_mutex or Free static signal id
	 * @param  none
	 * @return Delete success or not
	 * @retval true:Delete success
	 * @retval false:Delete failed
	 */
	bool Delete();

	/**
	 * @brief  Wait. If signal was not send, wait until get signal
	 * @param  none
	 * @return Get signal or not
	 * @retval true:Get signal
	 * @retval false:Not get signal
	 */
	bool Wait();

	/**
	 * @brief  Asyncronyze Wait. If signal was not send, return false
	 * @param  none
	 * @return Get signal or not
	 * @retval true:Get signal
	 * @retval false:Not get signal
	 */
	bool AsyncWait();

	/**
	 * @brief  Timed wait. If signal was not send, wait until times up
	 * @return Get signal or not
	 * @retval true:Get signal
	 * @retval false:Not get signal
	 */
	bool TimedWait(
		unsigned long timeout ///< [in] timeout time
	);

	/**
	 * @brief  Send signal
	 * @param  none
	 * @retval true:Get signal
	 * @retval false:Not get signal
	 */
	bool Send();

	/**
	 * @brief  Get error code. 
	 * @param  none
	 * @return Signal error code
	 * @retval SignalError
	 */
	SIGNAL_ERROR GetLastError();
}; // class Signal
#endif // defined(PAFEE_OS_WINDOWS) || defined(PAFEE_OS_LINUX) || defined(PAFEE_OS_ITRON)

}; // namespace Pafee

#endif // _PAFEE_SIGNAL_H_INCLUDED_
