//********************************************************************
//		project 	PAFEE
//
//! 	 @file: 	PafeeTask.cpp
//! 	 @author:	Technocrats
//! 	 @brief:	PafeeTask class
//
//********************************************************************
//
// P A F E E   - Pragmatic Application Framework for Embedded Environment- 
//
//
// Copyright (c) 2009 KN Information Sustems Corporation,
//                    Hirocon Corporation,
//                    Technocrats Corporation
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//

// if no OS definition, PafeeTask class is unavailable
#if defined(PAFEE_OS_WINDOWS)||defined(PAFEE_OS_LINUX)||defined(PAFEE_OS_ITRON)

#include <limits.h>
#include <stdarg.h>
#include "PafeeTask.h"
#include "PafeeTaskConfig.h"

#ifdef __cplusplus
namespace Pafee
{
#endif

#ifdef PAFEE_OS_ITRON

static struct
{
	int taskID;
	PafeeTask* pTask;
} g_taskTable[MAX_TASK] = {0, NULL};

#if defined(PAFEE_OS_ITRON)&&defined(PAFEE_DEBUG)
/**
 * @brief  Log output function(for ITRON simulator on Windows)
 * @param  *format, ... : Log contents
 * @return (none)
 */
void TaskSyslog(
	const char *format, ...
)
{
	const char *string;
	int data;

	va_list list;
	va_start(list, format);
	string = format;
	data = va_arg(list, int);
	syslog(LOG_NOTICE, format, data);
}
#endif // PAFEE_OS_ITRON && PAFEE_DEBUG

/**
 * @brief  global function to use thread procedure function (for ITRON)
 * @param  void* : Task procedure information structure
 * @return void pointer
 */
void TaskEntry(
	void* arg
)
{
	int taskNumber = 0;
	int count, place;
	TASK_ENTRY_PARAM* param;

	TASK_DEBUG((">> TaskEntry\n"));

	param = (TASK_ENTRY_PARAM*)arg;

	// register task ID with the global table
	for(count=0; count<MAX_TASK; count++)
	{
		if(g_taskTable[count].taskID == 0)
		{
			g_taskTable[count].taskID = param->taskID;
			TASK_DEBUG((">> PROC1 TaskEntry %d\n",param->taskID));
			place = count;
			break;
		}
	}
	TASK_DEBUG((">> PROC2 TaskEntry \n"));

	// check whether Task instance pointer registered with the global table
	while(g_taskTable[place].pTask == NULL)
	{
		TASK_DEBUG((">> PROC3 TaskEntry wait %d\n",g_taskTable[place].taskID));
		dly_tsk(1000);
	}
	(g_taskTable[place].pTask)->TaskProcedure(param->pTaskParam);

	TASK_DEBUG(("<< TaskEntry\n"));
}
#endif // PAFEE_OS_ITRON

/**
 * @brief  PafeeTask class constructor
 * @param  (none)
 * @return (none)
 */
PafeeTask::PafeeTask(
)
{
	TASK_DEBUG((">> PafeeTask::PafeeTask\n"));
#ifdef PAFEE_OS_WINDOWS
	m_handle = NULL;
#endif
#ifdef PAFEE_OS_LINUX
	m_handle = 0;
	m_isCreated = false;
#endif
#ifdef PAFEE_OS_ITRON
	m_handle = 0;
#endif
	m_error = ERROR_NONE;

#if 0
//#ifdef PAFEE_WINDOWS_VC6
	m_libraryHandle = NULL;
#endif
	TASK_DEBUG(("<< PafeeTask::PafeeTask\n"));
}

/**
 * @brief  PafeeTask class destructor
 * @param  (none)
 * @return (none)
 */
PafeeTask::~PafeeTask(
)
{
	TASK_DEBUG((">> PafeeTask::~PafeeTask\n"));
#if defined(PAFEE_OS_WINDOWS)||defined(PAFEE_OS_LINUX)
	Delete();
#endif
	TASK_DEBUG(("<< PafeeTask::~PafeeTask\n"));
}

#ifdef PAFEE_OS_WINDOWS
/**
 * @brief  Create new task (for Windows)
 * @param  TASK_ID* : Task ID
 * @param  TASK_PRIORITY : Task priority
 * @param  TASK_SCHDUKLING : Task scheduling (no effect)
 * @param  int : Stack size
 * @param  bool : Start task or not on creating
 * @return Create success or not
 * @retval true:create succeeded
 * @retval false:create failed
 */
bool PafeeTask::Create(
	TASK_ID* pTaskID,
	TASK_PRIORITY priority,
	TASK_SCHEDULING schedule,
	int stackSize,
	void* pParam,
	bool isStartImmediately
)
{
	TASK_ID retID;
	DWORD creationFlag;
	BOOL result;
	DWORD threadID;

	TASK_DEBUG((">> PafeeTask::Create\n"));

	if(m_handle != NULL)
	{
		m_error = ERROR_HAS_HANDLE;
		return false;
	}

#if 0
//#ifdef PAFEE_WINDOWS_VC6
	// load module library(for OpenThread())
	typedef HANDLE (WINAPI *MyProc)(DWORD, BOOL, DWORD);
	m_libraryHandle = LoadLibrary("kernel32");
	MyProc ProcAdd;
	if(m_libraryHandle != NULL)
	{
	    ProcAdd = (MyProc)GetProcAddress(m_libraryHandle, "OpenThread");
	}
#endif

	if(isStartImmediately == true)
	{
		creationFlag = 0;  // start the thread immediately
	}
	else
	{
		creationFlag = CREATE_SUSPENDED;
	}

	m_threadParam.instance = this;
	m_threadParam.pParam = pParam;
	retID = CreateThread(NULL, stackSize, TaskLauncher, &m_threadParam, creationFlag, &threadID);
	if(retID == NULL)
	{
		m_error = ERROR_CANNOT_CREATE;
		TASK_DEBUG(("<<FAIL1 PafeeTask::Create\n"));
		return false;  // error
	}

#if 0
//#ifdef PAFEE_WINDOWS_VC6
	retID = ProcAdd(SYNCHRONIZE|THREAD_ALL_ACCESS|THREAD_QUERY_INFORMATION|THREAD_SUSPEND_RESUME, TRUE, threadID);

	if(retID == NULL)
	{
		m_error = ERROR_CANNOT_CREATE;
		TASK_DEBUG(("<<FAIL2 PafeeTask::Create\n"));
		return false;  // error
	}
#endif

	m_handle = retID;  // keep retID for deleting case

	result = SetThreadPriority(retID, priority);
	if(result == 0)
	{
		Delete();
		m_error = ERROR_CANNOT_CREATE;
		TASK_DEBUG(("<<FAIL3 PafeeTask::Create\n"));
		return false;  // error
	}

	*pTaskID = retID;
	TASK_DEBUG(("<< PafeeTask::Create\n"));
	return true;  // succeeded in creating thread
}
#endif

#ifdef PAFEE_OS_LINUX
/**
 * @brief  Create new task (for Linux)
 * @param  TASK_ID* : Task ID
 * @param  TASK_PRIORITY : Task priority
 * @param  TASK_SCHEDULING : Task scheduling
 * @param  int : Stack size
 * @param  bool : Start task or not on creating (no effect)
 * @return Create success or not
 * @retval true:create succeeded
 * @retval false:create failed
 */
bool PafeeTask::Create(
	TASK_ID* pTaskID,
	TASK_PRIORITY priority,
	TASK_SCHEDULING schedule,
	int stackSize,
	void* pParam,
	bool isStartImmediately
)
{
	int result = 0;
	pthread_attr_t attribute;
	sched_param schedulePriority;

	TASK_DEBUG((">> PafeeTask::Create\n"));
	if(m_isCreated == true)
	{
		m_error = ERROR_HAS_HANDLE;
		TASK_DEBUG(("<<FAIL1 PafeeTask::Create\n"));
		return false;  // error
	}
	TASK_DEBUG((">>proc1 PafeeTask::Create\n"));

	result = pthread_attr_init(&attribute);
	if(result != 0)
	{
		m_error = ERROR_CANNOT_CREATE;
		TASK_DEBUG(("<<FAIL2 PafeeTask::Create\n"));
		return false;  // error
	}
	TASK_DEBUG((">>proc2 PafeeTask::Create\n"));

	if( (schedule == SCHED_RR)
	 || (schedule == SCHED_FIFO) )
	{
		if( (priority > sched_get_priority_max(schedule))
		 && (priority < sched_get_priority_min(schedule)) )
		{
			m_error = ERROR_CANNOT_CREATE;
			TASK_DEBUG(("<<FAIL3 PafeeTask::Create\n"));
			TASK_DEBUG(( "(priority = %d)\n",priority ));
			TASK_DEBUG(( "(priority_max = %d)\n",sched_get_priority_max(schedule) ));
			TASK_DEBUG(( "(priority_min = %d)\n",sched_get_priority_min(schedule) ));
			return false;  // error
		}
		else
		{
			schedulePriority.sched_priority = priority;
		}
	}
	TASK_DEBUG((">>proc3 PafeeTask::Create\n"));

	result = pthread_attr_setschedpolicy(&attribute, schedule);
	if(result != 0)
	{
		m_error = ERROR_CANNOT_CREATE;
		TASK_DEBUG(("<<FAIL4 PafeeTask::Create\n"));
		return false;  // error
	}
	TASK_DEBUG((">>proc4 PafeeTask::Create\n"));

	result = pthread_attr_setschedparam(&attribute, (const sched_param*)&schedulePriority); 
	if(result != 0)
	{
		m_error = ERROR_CANNOT_CREATE;
		TASK_DEBUG(("<<FAIL5 PafeeTask::Create\n"));
		return false;  // error
	}
	TASK_DEBUG((">>proc5 PafeeTask::Create\n"));

	if(stackSize < PTHREAD_STACK_MIN)
	{
		stackSize = PTHREAD_STACK_MIN;
	}
	result = pthread_attr_setstacksize(&attribute, (size_t)stackSize); 
	if(result != 0)
	{
		m_error = ERROR_CANNOT_CREATE;
		TASK_DEBUG(("<<FAIL6 PafeeTask::Create\n"));
		return false;  // error
	}
	TASK_DEBUG((">>proc6 PafeeTask::Create\n"));

	m_threadParam.instance = this;
	m_threadParam.pParam = pParam;
	result = pthread_create(pTaskID, &attribute, TaskLauncher, &m_threadParam); 
	if(result != 0)
	{
		m_error = ERROR_CANNOT_CREATE;
		TASK_DEBUG(("<<FAIL7 PafeeTask::Create\n"));
		return false;  // error
	}

	m_handle = *pTaskID;
	m_isCreated = true;
	TASK_DEBUG(("<< PafeeTask::Create\n"));
	return true;  // succeeded in creating task
}
#endif

#ifdef PAFEE_OS_ITRON
/**
 * @brief  Attach static task id (for ITRON)
 * @return Attach success or not
 * @retval true:Attach succeeded
 * @retval false:Attach failed
 */
bool PafeeTask::Attach(
	TASK_ID id ///< [in] Task ID
)
{
	int count;
	TASK_DEBUG((">> PafeeTask::Attach\n"));

	if(m_handle != 0)
	{
		m_error = ERROR_HAS_HANDLE;
		TASK_DEBUG(("<<FAIL1 PafeeTask::Attach\n"));
		return false;
	}

	while(m_handle == 0)  // wait for ID registered
	{
		TASK_DEBUG((">>proc1 PafeeTask::Attach waiting %d\n",id));
		dly_tsk(100);

		// register PafeeTask instance pointer with the global table
		for(count=0;count<MAX_TASK;count++)
		{
			if(g_taskTable[count].taskID == id)
			{
				m_handle = id;
				TASK_DEBUG((">>proc2 PafeeTask::Attach\n",id));
				g_taskTable[count].pTask = this;
				break;
			}
		}
	}

	TASK_DEBUG(("<< PafeeTask::Attach\n"));
	return true;
}
#endif

#ifdef PAFEE_OS_WINDOWS
/**
 * @brief  Start task(for Windows)
 * @param  (none)
 * @return Start success or not
 * @retval true:Start succeeded
 * @retval false:Start failed
 */
bool PafeeTask::Start(
)
{
	DWORD result;

	TASK_DEBUG((">> PafeeTask::Start\n"));
	if(m_handle == NULL)
	{
		m_error = ERROR_NOHANDLE;
		TASK_DEBUG(("<<FAIL1 PafeeTask::Start\n"));
		return false;  // error
	}
	else
	{
		result = ResumeThread(m_handle);
		// result :0  :the thread was already resumed
		//        :1  :the thread has just resumed
		//        :1+ :the thread keeps suspended
		//        :-1 :function failed

		if(result == -1)
		{
			m_error = ERROR_CANNOT_START;
			TASK_DEBUG(("<<FAIL2 PafeeTask::Start\n"));
			return false;  // error
		}
		else if(result > 1)
		{
			m_error = ERROR_STILL_SUSPENDED;
			TASK_DEBUG(("<<FAIL3 PafeeTask::Start\n"));
			return false;  // error
		}

	}
	TASK_DEBUG(("<< PafeeTask::Start\n"));
	return true;
}
#endif
#ifdef PAFEE_OS_ITRON
/**
 * @brief  Start task(for ITRON)
 * @param  (none)
 * @return Start success or not
 * @retval true:Start succeeded
 * @retval false:Start failed
 */
bool PafeeTask::Start(
)
{
	ER returnValue;
	TASK_DEBUG((">> PafeeTask::Start\n"));

	if(m_handle == 0)
	{
		m_error = ERROR_NOHANDLE;
		TASK_DEBUG(("<<FAIL1 PafeeTask::Start\n"));
		return false;  // error
	}
	else
	{
		returnValue = rsm_tsk(m_handle);
		if(returnValue != E_OK)
		{
			m_error = ERROR_CANNOT_START;
			TASK_DEBUG(("<<FAIL2 PafeeTask::Start %d\n",returnValue));
			return false;
		}

		TASK_DEBUG(("<< PafeeTask::Start\n"));
		return true;
	}
}
#endif

#ifdef PAFEE_OS_WINDOWS
/**
 * @brief  Stop task(for Windows)
 * @param  (none)
 * @return Stop success or not
 * @retval true:Stop succeeded
 * @retval false:Stop failed
 */
bool PafeeTask::Stop(
)
{
	DWORD result;

	TASK_DEBUG((">> PafeeTask::Stop\n"));
	if(m_handle == NULL)
	{
		m_error = ERROR_NOHANDLE;
		TASK_DEBUG(("<<FAIL1 PafeeTask::Stop\n"));
		return false;  // error
	}
	else
	{
		result = SuspendThread(m_handle);

		if(result == -1) // function failed
		{
			m_error = ERROR_CANNOT_STOP;
			TASK_DEBUG(("<<FAIL2 PafeeTask::Stop\n"));
			return false;  // error
		}
	}
	TASK_DEBUG(("<< PafeeTask::Stop\n"));
	return true;
}
#endif
#ifdef PAFEE_OS_ITRON
/**
 * @brief  Stop task(for ITRON)
 * @param  (none)
 * @return Stop success or not
 * @retval true:Stop succeeded
 * @retval false:Stop failed
 */
bool PafeeTask::Stop(
)
{
	ER returnValue;
	TASK_DEBUG((">> PafeeTask::Stop\n"));

	if(m_handle == 0)
	{
		m_error = ERROR_NOHANDLE;
		TASK_DEBUG(("<<FAIL1 PafeeTask::Start\n"));
		return false;  // error
	}
	else
	{
		returnValue = sus_tsk(m_handle);

		if(returnValue != E_OK)
		{
			m_error = ERROR_CANNOT_STOP;
			TASK_DEBUG(("<<FAIL2 PafeeTask::Stop %d\n",returnValue));
			return false;
		}

		TASK_DEBUG(("<< PafeeTask::Stop\n"));
		return true;
	}
}
#endif

#ifdef PAFEE_OS_WINDOWS
/**
 * @brief  Delete task(for Windows)
 * @param  (none)
 * @return Delete success or not
 * @retval true:Delete succeeded
 * @retval false:Delete failed
 */
bool PafeeTask::Delete(
)
{
	BOOL result;
	DWORD exitCode = 0;

	TASK_DEBUG((">> PafeeTask::Delete\n"));
	if(m_handle == NULL)
	{
		m_error = ERROR_NOHANDLE;
		TASK_DEBUG(("<<FAIL1 PafeeTask::Delete\n"));
		return false;  // error
	}

	result = GetExitCodeThread(m_handle, &exitCode);
	if(result == FALSE)
	{
		m_error = ERROR_CANNOT_DELETE;
		TASK_DEBUG(("<<FAIL2 PafeeTask::Delete\n"));
		return false;  // error
	}

	result = TerminateThread(m_handle, exitCode);
	if(result == FALSE)
	{
		m_error = ERROR_CANNOT_DELETE;
		TASK_DEBUG(("<<FAIL3 PafeeTask::Delete\n"));
		return false;  // error
	}

	if(m_handle != NULL)
	{
		CloseHandle(m_handle);
	}

#if 0
//#ifdef PAFEE_WINDOWS_VC6
	result = FreeLibrary(m_libraryHandle);
	if(result == FALSE)
	{
		m_error = ERROR_CANNOT_DELETE;
		TASK_DEBUG(("<<FAIL4 PafeeTask::Delete\n"));
		return false;  // error
	}
#endif

	m_handle = NULL;
	TASK_DEBUG(("<< PafeeTask::Delete\n"));
	return true;
}
#endif
#ifdef PAFEE_OS_LINUX
/**
 * @brief  Delete task(for Linux)
 * @param  (none)
 * @return Delete success or not
 * @retval true:Delete succeeded
 * @retval false:Delete failed
 */
bool PafeeTask::Delete(
)
{
	int result = 0;
	
	TASK_DEBUG((">> PafeeTask::Delete\n"));
	if(m_isCreated == false)
	{
		m_error = ERROR_NOHANDLE;
		TASK_DEBUG(("<<FAIL1 PafeeTask::Delete\n"));
		return false;  // error
	}

	if(m_handle == 0)
	{
		m_error = ERROR_NOHANDLE;
		TASK_DEBUG(("<<FAIL2 PafeeTask::Delete\n"));
		return false;  // error
	}

	result = pthread_detach(m_handle);
	if(result != 0)
	{
		m_error = ERROR_CANNOT_DELETE;
		TASK_DEBUG(("<<FAIL3 PafeeTask::Delete\n"));
		return false;  // error
	}

	m_isCreated = false;
	TASK_DEBUG(("<< PafeeTask::Delete\n"));
	return true;
}
#endif

/**
 * @brief  Task procedure function
 * @param  parameter
 * @return return value
 */
void* PafeeTask::TaskProcedure(
	void* pParam
)
{
	TASK_DEBUG((">> PafeeTask::TaskProcedure\n"));
	// inherit this function and implement procedure
	TASK_DEBUG(("<< PafeeTask::TaskProcedure\n"));
	return (void *)0;
}

/**
 * @brief  Get error code. 
 * @param  (none)
 * @return Task error code
 * @retval TaskError
 */
PafeeTask::TASK_ERROR PafeeTask::GetLastError(
)
{
	TASK_DEBUG((">> PafeeTask::GetLastError\n"));
	TASK_DEBUG(("<< PafeeTask::GetLastError\n"));
	return m_error;
}

#ifdef __cplusplus
}; // namespace pafee
#endif

#endif // PAFEE_OS_WINDOWS || PAFEE_OS_LINUX || PAFEE_OS_ITRON
