//****************************************************
//
//      project     PAFEE
//!      @file:     CommandOption.cpp
//!      @author:   Technocrats
//!      @brief:    Command-Line Option Analizer class
//
//*****************************************************

#include "StdAfx.h"
#include "define.h"
#include "assert.h"
#include "CommandOption.h"

using namespace std;

#ifdef __cplusplus
namespace Pafee
{
#endif

/**
 * @brief constractor of CommandOption Class
 */
CommandOption::CommandOption(
	const int argc,
	const char* argv[]
)
{
	m_configFilePath = ".\\pafeeConfig.ini";
	m_dependencyFilePath = ".\\pafeeConfig.depend";
	m_sourceTreePath = "..\\Src\\";
	m_outputPath = ".\\Output\\";
	m_isVersion = FALSE;
	m_isHelp = FALSE;
	m_isFreestanding = FALSE;
	m_isWrongInput = FALSE;
	AnalyzeOption(argc, (char **)argv);
}

/**
 * @brief destoractor of CommandOption Class
 */
CommandOption::~CommandOption(
)
{
}

/**
 * @brief  Get information from command line option
 * @param  argc : command line parameter
 * @param  argv[] : command line parameter
 * @return Anarysis success or not
 * @retval TRUE Anarysis success
 * @retval FALSE Anarysis failure
 */
BOOL CommandOption::AnalyzeOption(
	int argc,
	char* argv[]
)
{
	BOOL retval = FALSE;
	int count = 1;
	char *argp;

	while(count < argc)
	{
		argp = argv[count++];

		if(*argp == '-')
		{
			switch(*(++argp))
			{

			// -v : print version information
			case 'v':
				{
					if(*(++argp) == NULL)
					{
						m_isVersion = TRUE;
					}
					else
					{
						m_isWrongInput = TRUE;
					}
				}
				break;

			// -h : print help message
			case 'h':
				{
					if(*(++argp) == NULL)
					{
						m_isHelp = TRUE;
					}
					else
					{
						m_isWrongInput = TRUE;
					}
				}
				break;

			// -freestanding : enable C library subset
			case 'f':
				{
					if(strcmp("reestanding", ++argp) == 0)
					{
						m_isFreestanding = TRUE;
					}
					else
					{
						m_isWrongInput = TRUE;
					}
				}
				break;

			// -i <config file path> : set config file path
			case 'i':
				{
					if((*(++argp) == NULL)
					 &&(count < argc))
					{
						argp = argv[count++];
						m_configFilePath = argp;
					}
					else
					{
						m_isWrongInput = TRUE;
					}
				}
				break;

			// -d <dependency file path> : set dependency file path
			case 'd':
				{
					if((*(++argp) == NULL)
					 &&(count < argc))
					{
						argp = argv[count++];
						m_dependencyFilePath = argp;
					}
					else
					{
						m_isWrongInput = TRUE;
					}
				}
				break;

			// -o <output folder path> : set output folder path
			case 'o':
				{
					if((*(++argp) == NULL)
					 &&(count < argc))
					{
						argp = argv[count++];
						m_outputPath = argp;
					}
					else
					{
						m_isWrongInput = TRUE;
					}
				}
				break;

			// -s <source tree path> : set PAFEE source tree folder path
			case 's':
				{
					if((*(++argp) == NULL)
					 &&(count < argc))
					{
						argp = argv[count++];
						m_sourceTreePath = argp;
					}
					else
					{
						m_isWrongInput = TRUE;
					}
				}
				break;
			
			default:
				m_isWrongInput = TRUE;
				break;
			}
		}
		else
		{
			m_isWrongInput = TRUE;
		}
	}		// end of while

	return retval;
}

/**
 * @brief  Return Config File Path
 * @param  (none)
 * @return Config file path
 * @retval Config file path
 */
string CommandOption::GetConfigFilePath(
)
{
	return m_configFilePath;
}

/**
 * @brief  Return Dependency File Path
 * @param  (none)
 * @return Dependency file path
 * @retval Dependency file path
 */
string CommandOption::GetDependencyFilePath(
)
{
	return m_dependencyFilePath;
}

/**
 * @brief  Return Source Tree Folder Path
 * @param  (none)
 * @return Source Tree Folder Path
 * @retval Source Tree Folder Path
 */
string CommandOption::GetSourceTreePath(
)
{
	return m_sourceTreePath;
}

/**
 * @brief  Return Output Folder Path
 * @param  (none)
 * @return Output Folder Path
 * @retval Output Folder Path
 */
string CommandOption::GetOutputPath(
)
{
	return m_outputPath;
}

/**
 * @brief  Return whether Print Version information
 * @param  (none)
 * @return Print Version information or Not
 * @retval TRUE : Print Version information
 * @retval FALSE : Don't Print Version Information
 */
BOOL CommandOption::IsVersion(
)
{
	return m_isVersion;
}

/**
 * @brief  Return whether Print Help message
 * @param  (none)
 * @return Print Help message or Not
 * @retval TRUE : Print Help message
 * @retval FALSE : Don't Print Help message
 */
BOOL CommandOption::IsHelp(
)
{
	return m_isHelp;
}

/**
 * @brief  Return whether Use Library for Freestanding Environment
 * @param  (none)
 * @return Use Library for Freestanding Environment or Not
 * @retval TRUE : Use Library for Freestanding Environment
 * @retval FALSE : Don't Use Library for Freestanding Environment
 */
BOOL CommandOption::IsFreestanding(
)
{
	return m_isFreestanding;
}

/**
 * @brief  Return whether command line include wrong input
 * @param  (none)
 * @return command line include wrong input or Not
 * @retval TRUE : command line include wrong input
 * @retval FALSE : command line doesn't include wrong input
 */
BOOL CommandOption::IsWrongInput(
)
{
	return m_isWrongInput;
}


#ifdef __cplusplus
}; // namespace Pafee
#endif
