//****************************************************
//
//      project     PAFEE
//!      @file:     PafeeCrc.cpp
//!      @author:   chu
//!      @brief:    Crc class source file
//
//*****************************************************

//
// P A F E E   - Pragmatic Application Framework for Embedded Environment- 
//
//
// Copyright (c) 2009 KN Information Systems Corporation,
//                    Hirocon Corporation,
//                    Technocrats Corporation
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//

#include "PafeeCrc.h"

namespace Pafee
{

const unsigned long Crc::CRC_POLYNOMIAL = 0x04C11DB7;
const unsigned short Crc::CRC16_POLYNOMIAL = 0x1021;

//----------------------------------------------------------------------------
//!  Constructor
//!  @return
//----------------------------------------------------------------------------
Crc::Crc()
{
}

//----------------------------------------------------------------------------
//!  Destructor
//!  @return
//----------------------------------------------------------------------------
Crc::~Crc()
{
}

//----------------------------------------------------------------------------
//!  CRC16-CCITT's compute
//!  @return CRC16(CCITT) value
//----------------------------------------------------------------------------
unsigned short Crc::Encode16(
	void* pData, ///< [in] Point of data to be computed 
	int length ///< [in] Length of data in byte
)
{
	unsigned short crc16Table[256];
	pafee_memset(&crc16Table, 0, sizeof(crc16Table));

	// 256 values representing ASCII character codes.
	// compute table for CRC16-CCITT
	for(unsigned short i16Codes = 0; i16Codes <= 0xFF; i16Codes++)
	{
		crc16Table[i16Codes] = Reflect(i16Codes, 8) << 8;
		for(unsigned short iPos = 0; iPos < 8; iPos++)
		{
			crc16Table[i16Codes] = (crc16Table[i16Codes] << 1) ^
				(crc16Table[i16Codes] & (1 << 15) ? CRC16_POLYNOMIAL : 0);
		}
		crc16Table[i16Codes] = Reflect(crc16Table[i16Codes], 16);
	}
	unsigned char* p;
	p = (unsigned char*) pData;
	unsigned short crc16 = 0xffff;
	while(length--)
	{
		crc16 = (crc16 >> 8) ^ crc16Table[(crc16 ^ *p++) & 0xff];		
	}
	return ~crc16;
}

//----------------------------------------------------------------------------
//!  Crc-IEEE802.3's compute
//!  @return CRC32 value
//----------------------------------------------------------------------------
unsigned long Crc::Encode32(
	void* pData,  ///< [in] Point of data to be computed 
	int length  ///< [in] Length of data in byte
)
{
	unsigned long crc32Table[256];
	pafee_memset(&crc32Table, 0, sizeof(crc32Table));

	// 256 values representing ASCII character codes.
	// compute table for Crc-IEEE802.3
	for(unsigned long i32Codes = 0; i32Codes <= 0xFF; i32Codes++)
	{
		crc32Table[i32Codes] = Reflect(i32Codes, 8) << 24;
		for(unsigned long iPos = 0; iPos < 8; iPos++)
		{
			crc32Table[i32Codes] = (crc32Table[i32Codes] << 1) ^
				(crc32Table[i32Codes] & (1 << 31) ? CRC_POLYNOMIAL : 0);
		}

		crc32Table[i32Codes] = Reflect(crc32Table[i32Codes], 32);
	}
	unsigned char* p;
	p = (unsigned char*) pData;
	unsigned long Crc = 0xffffffff;
	while(length--)
	{
		Crc = (Crc >> 8) ^ crc32Table[(Crc & 0xFF) ^ *p++];
	}
	return ~Crc;
}

//----------------------------------------------------------------------------
//!  Reflect data
//!  @return Reflected data
//----------------------------------------------------------------------------
unsigned long Crc::Reflect(
	unsigned long data,        ///< [in] Data to be reflect
	unsigned char reflectSize  ///< [in] Reflect size
)
{
	unsigned long value = 0;
	// Swap bit 0 for bit 7 bit 1 For bit 6, etc....
	for(int iPos = 1; iPos < (reflectSize + 1); iPos++)
	{
		if(data & 1) 
		{
			value |= 1 << (reflectSize - iPos);
		}
		data >>= 1;
	}
	return value;
}

//----------------------------------------------------------------------------
//!  Reflect data
//!  @return Reflected data
//----------------------------------------------------------------------------
unsigned short Crc::Reflect(
	unsigned short data,        ///< [in] Data to be reflect
	unsigned char reflectSize  ///< [in] Reflect size
)
{
	unsigned short value = 0;
	// Swap bit 0 for bit 7 bit 1 For bit 6, etc....
	for(int iPos = 1; iPos < (reflectSize + 1); iPos++)
	{
		if(data & 1) 
		{
			value |= 1 << (reflectSize - iPos);
		}
		data >>= 1;
	}
	return value;
}

//----------------------------------------------------------------------------
//!  Verify CRC16-CCITT algorithm
//!  @return Result(true:right false:wrong)
//----------------------------------------------------------------------------
bool Crc::IsCrc16Good(
	void* pData, ///< [in] Point of data to be computed
	int length ///< [in] Length of data in byte
)
{
	unsigned short crc16Table[256];
	pafee_memset(&crc16Table, 0, sizeof(crc16Table));

	// 256 values representing ASCII character codes.
	// compute table for CRC16-CCITT
	for(unsigned short i16Codes = 0; i16Codes <= 0xFF; i16Codes++)
	{
		crc16Table[i16Codes] = Reflect(i16Codes, 8) << 8;
		for(unsigned short iPos = 0; iPos < 8; iPos++)
		{
			crc16Table[i16Codes] = (crc16Table[i16Codes] << 1) ^
				(crc16Table[i16Codes] & (1 << 15) ? CRC16_POLYNOMIAL : 0);
		}
		crc16Table[i16Codes] = Reflect(crc16Table[i16Codes], 16);
	}
	unsigned char* p;
	p = (unsigned char*) pData;
	unsigned short crc16 = 0xffff;
	while(length--)
	{
		crc16 = (crc16 >> 8) ^ crc16Table[(crc16 ^ *p++) & 0xff];
	}
	// 0xf0b8 us tge "Magic Value" of CRC-ITU
	if(crc16 == 0xf0b8)
	{
		return true;
		//cout << "true";
	}
	else
	{
		return false;
		//cout << "false";
	}
}

#ifdef PAFEE_UNITTEST
//----------------------------------------------------------------------------
//!	Unit test
//! @return 
//----------------------------------------------------------------------------
void Crc::UnitTest(void)
{
	Crc myCrc;
	unsigned short crc16;
	unsigned long Crc;
	char charEvenData1[4] = "ab";
	char charEvenData2[6] = "abcd";
	char charEvenData3[10] = "abcdefgh";
	char charEvenData4[18] = "abcdefghikjlmnop";
	char charEvenData5[34] = "abcdefghikjlmnopabcdefghikjlmnop";
	// 2bite, Test data:[ab]
	crc16 = Crc::Encode16(charEvenData1, 2);
	Crc = Crc::Encode32(charEvenData1, 2);
	// Encode32 test
	assert(Crc == 0x9e83486d);
	charEvenData1[2] = crc16 & 0xff;
	charEvenData1[3] = (crc16 >> 8) & 0xff;
	// Encode16 test
	assert(myCrc.IsCrc16Good(charEvenData1, 4));
	// 4bite, Test data:[abcd]
	crc16 = Crc::Encode16(charEvenData2, 4);
	Crc = Crc::Encode32(charEvenData2, 4);
	// Encode32 test
	assert(Crc == 0xed82cd11);
	charEvenData2[4] = crc16 & 0xff;
	charEvenData2[5] = (crc16 >> 8) & 0xff;	
	// Encode16 test
	assert(myCrc.IsCrc16Good(charEvenData2, 6));
	// 8bite, Test data:[abcdefgh]
	crc16 = Crc::Encode16(charEvenData3, 8);
	Crc = Crc::Encode32(charEvenData3, 8);
	// Encode32 test
	assert(Crc == 0xaeef2a50);
	charEvenData3[8] = crc16 & 0xff;
	charEvenData3[9] = (crc16 >> 8) & 0xff;
	// Encode16 test
	assert(myCrc.IsCrc16Good(charEvenData3, 10));
	// 16bite, Test data:[abcdefghikjlmnop]
	crc16 = Crc::Encode16(charEvenData4, 16);
	Crc = Crc::Encode32(charEvenData4, 16);
	// Encode32 test
	assert(Crc == 0xf9111882);
	charEvenData4[16] = crc16 & 0xff;
	charEvenData4[17] = (crc16 >> 8) & 0xff;
	// Encode16 test
	assert(myCrc.IsCrc16Good(charEvenData4, 18));
	// 32bite, Test data:[abcdefghikjlmnopabcdefghikjlmnop]
	crc16 = Crc::Encode16(charEvenData5, 32);
	Crc = Crc::Encode32(charEvenData5, 32);
	// Encode32 test
	assert(Crc == 0xcf110dbc);
	charEvenData5[32] = crc16 & 0xff;
	charEvenData5[33] = (crc16 >> 8) & 0xff;
	// Encode16 test
	assert(myCrc.IsCrc16Good(charEvenData5, 34));

	char charEvenData6[1000002];
	pafee_memset(charEvenData6,'a',1000000);

	crc16 = Crc::Encode16(charEvenData6, 1000000);
	Crc = Crc::Encode32(charEvenData6, 1000000);
	// Encode32 test
	assert(Crc == 0xdc25bfbc);
	charEvenData6[1000000] = crc16 & 0xff;
	charEvenData6[1000001] = (crc16 >> 8) & 0xff;
	// Encode16 test
	assert(myCrc.IsCrc16Good(charEvenData6, 1000002));
	
	unsigned long precrc16;
	unsigned long preCrc = 0UL;
	static int const MAXCOUNT =1000000;
	for(int i=1;i<MAXCOUNT;i=i+500)
	{
		char* charEvenData = new char[i+2];
		pafee_memset(charEvenData,'a',i);

		//Encode16,Encode32
		crc16 = Crc::Encode16(charEvenData, i);
		Crc = Crc::Encode32(charEvenData, i);
		// Encode32 test
		assert(preCrc != Crc);
		charEvenData[i] = crc16 & 0xff;
		charEvenData[i+1] = (crc16 >> 8) & 0xff;
		// Encode16 test
		assert(myCrc.IsCrc16Good(charEvenData, i+2));
		delete [] charEvenData;
		precrc16 = crc16;
		preCrc = Crc;
	}
}
#endif // PAFEE_UNITTEST

};	// Pafee
