//****************************************************
//
//      project     PAFEE
//!      @file:     PafeeZip.cpp
//!      @author:   chu
//!      @brief:    Zip class source file
//
//*****************************************************

//
// Copyright (C) 1995-2003 Jean-loup Gailly.
// For conditions of distribution and use, see copyright notice in zlib.h
//

//
// P A F E E   - Pragmatic Application Framework for Embedded Environment- 
//
//
// Copyright (c) 2009 KN Information Systems Corporation,
//                    Hirocon Corporation,
//                    Technocrats Corporation
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//

#include "PafeeZip.h"

namespace Pafee
{

//----------------------------------------------------------------------------
//!  Compresses the source buffer into the destination buffer using GZip
//!  @return 
//!  Z_OK: success, 
//!  Z_MEM_ERROR: there was not enough memory, 
//!  Z_BUF_ERROR: there was not enough room in the output buffer 
//----------------------------------------------------------------------------
bool Zip::Encode(
                 void* pDestination,        ///< [out] destination buffer
                 unsigned long* pDestinationLength, ///< [out] the total size of the destination buffer 
                 const void* pSource,       ///< [in] source buffer
                 unsigned long sourceLength         ///< [in] the byte length of the source buffer
                )
{
	Byte* pDestinationBuffer;
	Byte* pSourceBuffer;
	pDestinationBuffer = (Byte*) pDestination;
	pSourceBuffer = (Byte*) pSource;
	return compress(pDestinationBuffer, pDestinationLength, pSourceBuffer, sourceLength) == Z_OK;
}

//----------------------------------------------------------------------------
//!  Decompresses the source buffer into the destination buffer using GZip
//!  @return
//!  Z_OK: success, 
//!  Z_MEM_ERROR: there was not enough memory, 
//!  Z_BUF_ERROR: there was not enough room in the output buffer, 
//!  Z_DATA_ERROR: the input data was corrupted or incomplete
//----------------------------------------------------------------------------
bool Zip::Decode(
                 void* pDestination,        ///< [out] destination buffer
                 unsigned long* pDestinationLength, ///< [out] the total size of the destination buffer 
                 const void* pSource,       ///< [in] source buffer
                 unsigned long sourceLength         ///< [in] the byte length of the source buffer
                )
{
	Byte* pDestinationBuffer;
	Byte* pSourceBuffer;
	pDestinationBuffer = (Byte*) pDestination;
	pSourceBuffer = (Byte*) pSource;
	return uncompress(pDestinationBuffer, pDestinationLength, pSourceBuffer, sourceLength) == Z_OK;
}

#ifdef PAFEE_UNITTEST
//----------------------------------------------------------------------------
//! Unit test
//! @return void
//----------------------------------------------------------------------------
void Zip::UnitTest(void)
{
	void *compr, *uncompr;
	unsigned long comprLen;
	unsigned long uncomprLen;
	Zip zip;
	unsigned long len;

	//test pattern 1
	const char hello[] = "hello, hello!";
//	comprLen = 10000*sizeof(int); ///< don't overflow on MSDOS
	comprLen = 10000;
	uncomprLen = comprLen;
	compr    = (void*)calloc((uInt)comprLen, 1);
	uncompr  = (void*)calloc((uInt)uncomprLen, 1);
	if (compr == NULL || uncompr == NULL)
	{
		printf("out of memory\n");
		exit(1);
	}
	len = (unsigned long)pafee_strlen(hello)+1;
	// Test compress()
	assert(zip.Encode(compr, &comprLen, (const void*)hello, len) == true);
	// Test uncompress()
	assert(zip.Decode(uncompr, &uncomprLen, compr, comprLen) == true );
	// Test 
	assert(pafee_memcmp(hello,(char*)uncompr,len) == 0);
	free(compr);
	free(uncompr);

	//test pattern 2
	char hello2[1000000];
	pafee_memset(hello2,'a',1000000);
	comprLen = 1000000;
	uncomprLen = comprLen;
	compr    = (void*)calloc((uInt)comprLen, 1);
	uncompr  = (void*)calloc((uInt)uncomprLen, 1);
	if (compr == NULL || uncompr == NULL)
	{
		printf("out of memory\n");
		exit(1);
	}
	len = (unsigned long)sizeof(hello2);
	// Test compress()
	assert(zip.Encode(compr, &comprLen, (const void*)hello2, len) == true);
	// Test uncompress()
	assert(zip.Decode(uncompr, &uncomprLen, compr, comprLen) == true );
	// Test 
	assert(pafee_memcmp(hello2,(char*)uncompr,len) == 0);
	free(compr);
	free(uncompr);

	//test pattern 3
	len = 1000000;
	static int hello3[1000000];
	for(unsigned int i=0;i<len;i++)
	{
		hello3[i] = rand();
	}
	comprLen = len;
	uncomprLen = comprLen;
	compr    = (void*)calloc((uInt)comprLen, 1);
	uncompr  = (void*)calloc((uInt)uncomprLen, 1);
	if (compr == NULL || uncompr == NULL)
	{
		printf("out of memory\n");
		exit(1);
	}
	// Test compress()
	assert(zip.Encode(compr, &comprLen, (const void*)hello3, len) == true);
	// Test uncompress()
	assert(zip.Decode(uncompr, &uncomprLen, compr, comprLen) == true );
	// Test 
	assert(pafee_memcmp(hello3,uncompr,len) == 0);
	free(compr);
	free(uncompr);

}

#endif

};	// Pafee
